# ruff: noqa: E501
# This file has example strings in templatize() that are intentionally formatted
import copy
import itertools
import json
import logging
import pathlib
import re
import uuid
from typing import TYPE_CHECKING, Callable, Union

import numpy as np

from rhesis.sdk.adaptive_testing.schemas import TestTreeData, TestTreeNode

# from ._scorer import expand_template, clean_template, Scorer
from .comm import JupyterComm
from .generators import Generator

if TYPE_CHECKING:
    from rhesis.sdk.models import BaseEmbedder

    from ._prompt_builder import PromptBuilder
    from .embedders import EmbedderAdapter

log = logging.getLogger(__name__)


# import sys
# sys.stderr = open('/tmp/err.txt', 'w')


def expand_template(s, keep_braces=False):
    """Expand a template string into a list of strings."""
    # parts = []
    # for s in strings:
    matches = re.findall("{[^}]*}", s)
    s = re.sub("{[^}]*}", "{}", s)
    template_groups = [str(m)[1:-1].split("|") for m in matches]
    try:
        if keep_braces:
            return [
                s.format(*["{{{p}}}" for p in parts])
                for parts in itertools.product(*template_groups)
            ]
        else:
            return [s.format(*parts) for parts in itertools.product(*template_groups)]
    except ValueError:
        return [s]  # we return the template not filled in if it is invalid


def clean_template(s):
    """This removes duplicate template entries."""
    matches = re.findall("{[^}]*}", s)
    s = re.sub("{[^}]*}", "{}", s)
    template_groups = [str(m)[1:-1].split("|") for m in matches]
    clean_groups = [
        "{" + "|".join(list({v: None for v in g}.keys())) + "}" for g in template_groups
    ]
    try:
        return s.format(*clean_groups)
    except ValueError:
        return s  # we return the template not cleaned in if it is invalid


def matches_filter(test, filter_text):
    if filter_text is None or filter_text == "":
        return True
    else:
        return filter_text in test.input or filter_text in test.output


valid_comparators = ["should not be", "should be", "should be the same as for"]
FILLIN_PREFIX = "/Fill-ins"


# model("this is english") => []
# output_sampling="topk(10)"
# output_sampling="topp(10)"
# output_sampling="max"
# output_sampling="temperature(0.9)"


class TestTreeBrowser:
    """Used for browsing and expanding a test tree."""

    def __init__(
        self,
        test_tree: "TestTreeData",
        generator: Generator,
        endpoint: Callable[[list[str]], list[str]],
        metrics: Callable[[list[str], list[str]], list[float]],
        embedder: Union["BaseEmbedder", "EmbedderAdapter", None],
        user: str,
        recompute_scores: bool,
        regenerate_outputs: bool,
        max_suggestions: int,
        prompt_variants: int,
        prompt_builder: "PromptBuilder",
        starting_path: str,
        score_filter: Union[float, str],
    ) -> None:
        """Initialize the TestTreeBrowser."""
        from rhesis.sdk.models import BaseEmbedder

        from .embedders import EmbedderAdapter

        self.test_tree = test_tree._tests
        self.topic_tree = self.test_tree.topics

        self.endpoint = endpoint
        self.metrics = metrics
        self.generator = generator
        self.user = user
        self.recompute_scores = recompute_scores
        self.regenerate_outputs = regenerate_outputs
        self.max_suggestions = max_suggestions
        self.prompt_variants = prompt_variants
        self.prompt_builder = prompt_builder
        self.current_topic = starting_path
        self.score_filter = score_filter
        self.filter_text = ""
        self._id = uuid.uuid4().hex
        self.mode = "tests"  # default mode: "tests" or "topics"

        # Set up embedder - wrap BaseEmbedder in adapter for caching compatibility
        if embedder is not None:
            if isinstance(embedder, BaseEmbedder):
                self.embedder = EmbedderAdapter(embedder)
            else:
                # Already an adapter or compatible object
                self.embedder = embedder
        else:
            raise ValueError("Embedder is required")

        # if we are recomputing the scores then we erase all the old scores
        # Reset scores if recomputing
        if recompute_scores is True:
            for node in self.test_tree:
                node.model_score = float("nan")

        # if regenerating outputs, force all tests to be re-evaluated
        if regenerate_outputs is True:
            for node in self.test_tree:
                if node.label != "topic_marker":
                    node.to_eval = True

        # these are all temporary state
        self._hidden_topics = {}
        self.comm = None

        # apply all the scorers to the test tree (this updates the test tree)
        # When regenerate_outputs=True, overwrite existing outputs with fresh ones from endpoint
        self._compute_embeddings_and_scores(
            recompute=self.recompute_scores,
            overwrite_outputs=self.regenerate_outputs,
            save_outputs=not self.regenerate_outputs,
        )

        # # make sure all the tests have scores (if we have a scorer)
        # self._compute_embeddings_and_scores(self.test_tree)

        # ensure test tree based generator has embeddings calculated
        if getattr(self.generator, "gen_type", "") == "test_tree":
            self.generator.source._cache_embeddings(embedder=self.embedder)

        # init a blank set of suggetions
        self._suggestions_error = ""  # tracks if we failed to generate suggestions

    def embed(self, strings):
        """Embed strings using the configured embedder."""
        from .embedders import embed_with_cache

        return embed_with_cache(self.embedder, strings)

    def _repr_html_(self, prefix="", environment="jupyter", websocket_server=None):
        """Returns the HTML interface for this browser.

        Parameters
        ----------
        prefix : str
            The URL prefix this test tree browser is being served from.

        environment : str
            The environment this test tree browser is being served from (jupyter or web).
        """

        # spin up a JupyterComm object if we are called directly (which we assume is in a notebook)
        if self.comm is None and environment == "jupyter":
            self.comm = JupyterComm(f"adatest_interface_target_{self._id}", self.interface_event)

        # dump the client javascript to the interface
        file_path = pathlib.Path(__file__).parent.absolute()
        with open(file_path / "resources" / "main.js", encoding="utf-8") as f:
            js_data = f.read()
        interface_html = f"""
<div id="adatest_container_{self._id}" style="width: 100%; all: initial;"></div>
<script type='text/javascript'>
  {js_data};
  AdaTestReactDOM.render(
    AdaTestReact.createElement(AdaTest, {{
      interfaceId: "{self._id}",
      environment: "{environment}",
      startingTopic: "{self.current_topic}",
      prefix: "{prefix}",
      websocket_server: {"undefined" if websocket_server is None else '"' + websocket_server + '"'},
    }}, null),
    document.getElementById('adatest_container_{self._id}')
  );
</script>
"""
        return interface_html

    def interface_event(self, msg):
        """Handle interface events from the client.

        Parameters
        ----------
        msg : dict
            The event messages from the client. Each key in the dictionary is a
            separate message to either the row specified by the key or to whole
            browser object if the key is 'browser'.
        """

        log.debug(f"interface_event({msg})")

        if "event_id" not in msg:
            log.error("interface_event: missing event_id. msg dump: %s", msg)
            return
        event_id = msg["event_id"]

        # redraw the entire interface
        if event_id == "redraw":
            self._refresh_interface()

        # generate a new set of suggested tests/topics
        elif event_id == "generate_suggestions":
            self._clear_suggestions()
            self._generate_suggestions(filter=msg.get("filter", ""))
            # if self._active_generator_obj is None:
            #     self._suggestions_error = "No adaptive_testing generator has been set!"
            # else:
            #     self._generate_suggestions(filter=msg[k].get("filter", ""))
            # # try:
            # self.suggestions = self._generate_suggestions(filter=msg[k].get("filter", ""))
            # # filter suggestions to relevant types
            # if self.mode == "topics":
            #     self.suggestions = self.suggestions[self.suggestions['type'] == "topic_marker"]
            # elif self.mode == "tests":
            #     self.suggestions = self.suggestions[self.suggestions['type'] != "topic_marker"]

            # # Ensure valid suggestions exist.
            # if self.suggestions.shape[0] > 0:
            #     self.suggestions.sort_values(
            #         self.score_columns[0], inplace=True, ascending=False,
            #         key=np.vectorize(score_max)
            #     )
            #     self._suggestions_error = ""
            # else:
            #     self._suggestions_error = True # Not sure if we should do this?
            # except Exception as e:
            #     log.debug(e)
            #     self.suggestions = pd.DataFrame([], columns=self.test_tree.columns)
            #     self._suggestions_error = True
            self._refresh_interface()

        # change the current topic (navigate to a different topic)
        elif event_id == "change_topic":
            self.current_topic = msg["topic"].lstrip("/")  # Remove leading slash
            self._refresh_interface()

        # clear the current set of suggestions
        elif event_id == "clear_suggestions":
            self._clear_suggestions()
            # self.suggestions = pd.DataFrame([], columns=self.test_tree.columns)
            self._refresh_interface()

        # add a new empty subtopic to the current topic
        elif event_id == "add_new_topic":
            new_topic_path = (
                self.current_topic + "/New topic" if self.current_topic else "New topic"
            )
            self.topic_tree.create(new_topic_path, labeler=self.user)
            self._refresh_interface()

        # add a new empty test to the current topic
        elif event_id == "add_new_test":
            # add the new test row (special value "New test" causes interface to auto-select)
            node = TestTreeNode(
                topic=self.current_topic,
                input="New test",
                output="",
                label="",
                labeler="imputed",
            )
            self.test_tree[node.id] = node

            self._refresh_interface()

        # change which scorer/model is used for sorting tests
        # NOTE: Currently only single model_score supported, this is a no-op
        elif event_id == "set_first_model":
            self._refresh_interface()

        elif event_id == "change_mode":
            self.mode = msg["mode"]

        elif event_id == "change_filter":
            print("change_filter")
            self.filter_text = msg["filter_text"]
            self._refresh_interface()

        # Move a test/topic to a new path (also used for renaming)
        elif event_id == "move_test":
            log.debug(f"move_test: test_ids={msg['test_ids']}, topic={msg['topic']}")
            test_ids = msg["test_ids"]
            new_path = msg["topic"].lstrip("/")  # Remove leading slash if present

            for test_id in test_ids:
                if test_id in self.test_tree.index:
                    # It's a node UUID - move that single node to the new topic
                    self.test_tree[test_id].topic = new_path
                    log.debug(f"  Moved node {test_id} to {new_path}")
                else:
                    # It's a topic path - move/rename the entire topic
                    topic = self.topic_tree.get(test_id)
                    if topic:
                        self.topic_tree.move(topic, new_path)
                        log.debug(f"  Moved topic {test_id} to {new_path}")

            self._compute_embeddings_and_scores()
            self._refresh_interface()

        elif event_id == "delete_test":
            log.debug("delete_test")
            test_ids = msg["test_ids"]
            # test_id can either be a unique test ID or a topic path
            for test_id in test_ids:
                if test_id in self.test_tree.index:
                    # It's a node UUID - delete that single node
                    del self.test_tree._nodes[test_id]
                else:
                    # It's a topic path - delete the topic and all its contents
                    topic = self.topic_tree.get(test_id)
                    if topic:
                        self.topic_tree.delete(topic, move_tests_to_parent=False)
            self._compute_embeddings_and_scores()
            self._refresh_interface()

        # if we are just updating a single row in tests then we only recompute the scores
        elif (
            event_id == "change_label" or event_id == "change_input" or event_id == "change_output"
        ):
            sendback_data = {}
            test_id = msg["test_ids"][0]

            # update the node fields and recompute scores
            node = self.test_tree[test_id]

            # convert template expansions into a standard value update
            if msg.get("action", "") == "template_expand":
                template_value = self.templatize(getattr(node, msg["value"]))
                msg = {msg["value"]: template_value}
                sendback_data[msg["value"]] = template_value

            metadata_fields = ["event_id", "test_ids"]
            for k2 in msg:
                if k2 not in metadata_fields and hasattr(node, k2):
                    setattr(node, k2, msg[k2])
            if event_id == "change_input":
                node.model_score = float("nan")
                node.to_eval = True
                self._compute_embeddings_and_scores(overwrite_outputs="output" not in msg)
            elif event_id == "change_label":
                # sign = -1 if msg["label"] == "pass" else 1
                # self.test_tree.loc[test_id, self.score_columns] = (
                #     abs(float(self.test_tree.loc[test_id, self.score_columns])) * sign
                # )
                # SML: we could recompute the scores here but then that would change
                # the output of stochastic output models
                pass
                # self._compute_embeddings_and_scores(self.test_tree, overwrite_outputs=False)

            # send just the data that changed back to the frontend
            sendback_data["scores"] = {
                "model_score": [[test_id, v] for v in ui_score_parts(node.model_score, node.label)]
            }
            # if the output was given to us the client is managing its current state
            # so we shouldn't send it back
            if "output" not in msg:
                sendback_data["output"] = node.output
            sendback_data["label"] = node.label
            sendback_data["labeler"] = node.labeler
            sendback_data.update(self.test_display_parts(node))
            self.comm.send({test_id: sendback_data})

        else:
            log.error(f"Unable to parse the interface message: {msg}")

    def _refresh_interface(self):
        """Send our entire current state to the frontend interface."""

        data = {}

        # Build data for current topic
        children = self._build_topic_children(data, self.current_topic)

        # Build data for suggestions
        suggestions_topic = (
            self.current_topic + "/__suggestions__" if self.current_topic else "__suggestions__"
        )
        suggestions_children = self._build_topic_children(data, suggestions_topic)

        # TODO: This is a complete hack to hide lower scoring suggestions when we are
        # likely already in the exploit phase
        # this is just for users who don't know when to stop scrolling down...
        # SML: I expect we can delete this at some point?
        if self.score_filter == "auto":
            if len(children) < 10:
                score_filter = -1e12
            else:
                children_scores = sorted(
                    [
                        np.max([score_max(x[1]) for x in data[key]["scores"]["model_score"]])
                        for key in children
                    ]
                )
                suggestions_children_scores = sorted(
                    [
                        np.max([score_max(x[1]) for x in data[key]["scores"]["model_score"]])
                        for key in suggestions_children
                    ]
                )
                score_filter = (
                    children_scores[-5] - (children_scores[-1] - children_scores[-5]) * 0.2
                )
                if len(suggestions_children_scores) > 0:
                    score_filter = min(score_filter, np.nanmax(suggestions_children_scores) - 1e-2)
        else:
            score_filter = self.score_filter

        # Compute mode based on current topic structure
        # "topics" mode: Generate creates subtopics; "tests" mode: Generate creates tests
        current = self.topic_tree.get(self.current_topic) if self.current_topic else None
        if current:
            has_direct_tests = self.topic_tree.has_direct_tests(current)
            has_subtopics = self.topic_tree.has_subtopics(current)
        else:
            # At root level
            has_direct_tests = False
            has_subtopics = len(self.topic_tree.get_children(None)) > 0
        self.mode = "topics" if not has_direct_tests and has_subtopics else "tests"

        topic_marker_id = self._get_topic_marker_id(self.current_topic)
        # compile the global browser state for the frontend
        data["browser"] = {
            "suggestions": suggestions_children,
            "tests": children,
            "user": self.user,
            "topic": self.current_topic,
            "topic_marker_id": topic_marker_id if topic_marker_id is not None else uuid.uuid4().hex,
            "score_filter": score_filter,
            "disable_suggestions": False,
            "read_only": False,
            "score_columns": ["model_score"],
            "suggestions_error": self._suggestions_error,
            "mode": self.mode,
            "test_tree_name": "Tests",
            # "test_types": test_types,
            # "test_type_parts": test_type_parts,
        }

        self.comm.send(data)

    def _build_topic_children(self, data: dict, topic_path: str) -> list[str]:
        """Build UI data for children of a topic and return sorted child IDs.

        Args:
            data: Dictionary to populate with UI data (modified in place)
            topic_path: The topic path to get children for

        Returns:
            Sorted list of child IDs (topic paths for folders, node IDs for tests)
        """
        children = []

        # Get child topics (folders)
        child_topics = self._get_child_topics_for_path(topic_path)
        for child_topic in child_topics:
            marker_id = self.topic_tree.get_topic_marker_id(child_topic)

            # Get all tests under this topic for aggregate scoring
            all_tests = self.topic_tree.get_tests(child_topic, recursive=True)
            scores = []
            for test in all_tests:
                scores.extend([[test.id, v] for v in ui_score_parts(test.model_score, test.label)])

            data[child_topic.path] = {
                "label": "topic_marker",
                "labeler": "",
                "scores": {"model_score": scores},
                "topic_marker_id": marker_id,
                "topic_name": child_topic.name,
                "editing": child_topic.name == "New topic",
            }
            children.append(child_topic.path)

        # Get direct tests (not in subtopics)
        direct_tests = self._get_direct_tests_for_path(topic_path)
        for test in direct_tests:
            if not matches_filter(test, self.filter_text):
                continue

            data[test.id] = {
                "input": test.input,
                "output": test.output,
                "label": test.label,
                "labeler": test.labeler,
                "scores": {
                    "model_score": [
                        [test.id, v] for v in ui_score_parts(test.model_score, test.label)
                    ]
                },
                "editing": test.input == "New test",
            }
            data[test.id].update(self.test_display_parts(test))
            children.append(test.id)

        # Sort children
        return self._sort_children(children, data)

    def _get_child_topics_for_path(self, topic_path: str) -> list:
        """Get child topics for a path, handling __suggestions__ pseudo-topics."""
        from .schemas import TopicNode

        if "__suggestions__" in topic_path:
            # For suggestions, manually find child topic markers
            child_topics = []
            prefix = topic_path + "/"
            seen = set()
            for node in self.test_tree:
                if node.label == "topic_marker" and node.topic.startswith(prefix):
                    # Extract direct child only
                    remainder = node.topic[len(prefix) :]
                    if "/" not in remainder and node.topic not in seen:
                        seen.add(node.topic)
                        child_topics.append(TopicNode(path=node.topic))
            return child_topics
        else:
            # Use TopicTree for regular topics
            topic = self.topic_tree.get(topic_path) if topic_path else None
            return self.topic_tree.get_children(topic)

    def _get_direct_tests_for_path(self, topic_path: str) -> list:
        """Get direct tests for a path, handling __suggestions__ pseudo-topics."""
        if "__suggestions__" in topic_path:
            # For suggestions, manually find direct tests
            return [
                node
                for node in self.test_tree
                if node.topic == topic_path and node.label != "topic_marker"
            ]
        else:
            # Use TopicTree for regular topics
            topic = self.topic_tree.get(topic_path) if topic_path else None
            if topic:
                return self.topic_tree.get_tests(topic, recursive=False)
            else:
                # Root level - get tests with empty topic
                return [
                    node
                    for node in self.test_tree
                    if node.topic == "" and node.label != "topic_marker"
                ]

    def _sort_children(self, children: list[str], data: dict) -> list[str]:
        """Sort children by score and type."""

        def sort_key(id):
            try:
                total = 0
                count = 0
                for s in data[id]["scores"]["model_score"]:
                    val = score_max(s[1], nan_val=np.nan)
                    if not np.isnan(val) and val is not None:
                        total += val
                        count += 1
                if count == 0:
                    return 1e3
                else:
                    return -total / count
            except Exception as e:
                log.warning(f"Error sorting {id}: {e}")
                return 1e3

        # Sort by score
        sorted_children = sorted(children, key=sort_key)
        # Put folders first
        sorted_children = sorted(
            sorted_children,
            key=lambda id: 0 if data[id].get("label", "") == "topic_marker" else 1,
        )
        # Put off_topic last
        sorted_children = sorted(
            sorted_children,
            key=lambda id: 1 if data[id].get("label", "") == "off_topic" else 0,
        )
        # Put new items first
        sorted_children = sorted(
            sorted_children,
            key=lambda id: (
                0
                if id.endswith("/New topic")
                or id == "New topic"
                or data[id].get("value1", "") == "New test"
                else 1
            ),
        )

        return sorted_children

    def _clear_suggestions(self):
        """Clear the suggestions for the current topic."""
        suggestions_prefix = (
            self.current_topic + "/__suggestions__" if self.current_topic else "__suggestions__"
        )
        ids_to_delete = [
            node.id for node in self.test_tree if node.topic.startswith(suggestions_prefix)
        ]
        for node_id in ids_to_delete:
            del self.test_tree._nodes[node_id]

    def generate_suggestions(self, topic=None, filter=""):
        if topic is not None:
            self.current_topic = topic
        self._clear_suggestions()
        self._generate_suggestions(filter=filter)

    def _generate_suggestions(self, filter):
        """Generate suggestions for the current topic.

        Parameters
        ----------
        filter : str
            The filter to apply to the tests while generating suggestions.
        """

        # --Backend-driven suggestions--

        # save a lookup we can use to detect duplicate tests
        test_map = {
            f"{test.topic} __topic_marker__"
            if test.label == "topic_marker"
            else f"{test.topic} __JOIN__ {test.input}"
            for test in self.test_tree
        }

        # validity focused (focus first on making valid in-topic tests,
        #   then secondarily on making those tests high scoring)
        # failure focused (focus on making high scoring (failing) tests,
        #   then secondarily on making those tests valid and in-topic)
        # topics (suggest new sub-topics)
        # file_name dataset (suggest tests based on samples from the provided dataset)

        # generate the prompts for the backend
        prompts = self.prompt_builder(
            test_tree=self.test_tree,
            topic=self.current_topic,
            score_column="model_score",
            repetitions=self.prompt_variants,
            filter=filter,
            suggest_topics=self.mode == "topics",
            embed_fn=self.embed,
        )

        # generate the suggestions
        proposals = self.generator(
            prompts,
            self.current_topic,
            self.mode,
            num_samples=self.max_suggestions // len(prompts)
            if len(prompts) > 0
            else self.max_suggestions,
        )

        if self.mode == "topics":
            proposals = list(proposals)

        # Build up suggestions catalog
        if True:  # Always use node-based approach
            # suggestions = []
            test_map_tmp = copy.copy(test_map)
            for input in proposals:
                if self.mode == "topics" and ("/" in input or "\n" in input):
                    input = input.replace("/", " or ").replace(
                        "\n", " "
                    )  # topics can't have newlines or slashes in their names
                    input = input.replace(
                        "  ", " "
                    ).strip()  # kill any double spaces we may have introduced
                    str_val = self.current_topic + "/" + input + " __topic_marker__"
                else:
                    str_val = self.current_topic + " __JOIN__ " + input
                if str_val not in test_map_tmp:
                    # Build suggestion topic path
                    if self.current_topic:
                        suggestion_topic = (
                            self.current_topic
                            + "/__suggestions__"
                            + ("/" + input if self.mode == "topics" else "")
                        )
                    else:
                        # Root level suggestions
                        suggestion_topic = "__suggestions__" + (
                            "/" + input if self.mode == "topics" else ""
                        )
                    node = TestTreeNode(
                        topic=suggestion_topic,
                        input="" if self.mode == "topics" else input,
                        output="[no output]",
                        label="topic_marker" if self.mode == "topics" else "",
                        labeler="imputed",
                        model_score=float("nan"),
                        to_eval=True,
                    )
                    self.test_tree[node.id] = node

                    # s = {
                    #     "topic": self.current_topic + "/__suggestions__"
                    #         + ("/" + input if self.mode == "topics" else ""),
                    #     "input": "" if self.mode == "topics" else input,
                    #     "output": "",
                    #     "label": "",
                    #     "labeler": "imputed",
                    #     "description": ""
                    # }
                    # for c in self.score_columns:
                    #     s[c] = ""
                    # suggestions.append(s)
                    if str_val is not None:
                        test_map_tmp.add(str_val)

            # suggestions = pd.DataFrame(
            #     suggestions, index=[uuid.uuid4().hex for _ in range(len(suggestions))],
            #     columns=self.test_tree.columns
            # )

            # compute the scores for the new tests
            self._compute_embeddings_and_scores()

    def _get_topic_marker_id(self, topic_path: str) -> str | None:
        """
        Returns the id of the topic marker row for the given topic path.
        Returns None if not found.
        """
        topic = self.topic_tree.get(topic_path)
        if topic:
            return self.topic_tree.get_topic_marker_id(topic)
        return None

    def _compute_embeddings_and_scores(
        self, recompute=False, overwrite_outputs=False, save_outputs=False
    ):
        log.debug(
            f"compute_embeddings_and_scores(tests=<TestTreeData len={len(self.test_tree)}>, "
            f"recompute={recompute})"
        )

        eval_ids = [
            node.id for node in self.test_tree if node.to_eval and node.label != "topic_marker"
        ]

        if len(eval_ids) > 0:
            # run the scorer

            eval_inputs = []
            eval_inds = []
            for i, id in enumerate(eval_ids):
                test = self.test_tree[id]
                template_expansions = expand_template(test.input)
                for expansion in template_expansions:
                    eval_inputs.append(expansion)
                    eval_inds.append(i)

            new_outputs = self.endpoint(eval_inputs)

            scores = self.metrics(eval_inputs, new_outputs)

            # update the scores in the test tree
            for i, id in enumerate(eval_ids):
                node = self.test_tree[id]
                # skip topic markers - they shouldn't have outputs or scores
                if node.label == "topic_marker":
                    continue
                node.output = new_outputs[i]
                node.model_score = scores[i]
                node.label = "fail" if scores[i] > 0.5 else "pass"
                node.labeler = "scored"
                node.to_eval = False

    def test_display_parts(self, test):
        # # find which template instantiation has the highest score (and so should be displayed)
        # ui_score_parts = test[self.score_columns[0]].split("|")
        # if len(ui_score_parts) == 1:
        #     max_score_ind = 0
        # else:
        #     max_score_ind = np.argmax([float(v) for v in test[self.score_columns[0]].split("|")])

        # if test.type == "{} should not be completed to become {}":
        #     value1_outputs = safe_json_load(
        #         test.get(self.score_columns[0][:-6] + " value1 outputs", "{}")
        #     ).get("string", [""])
        #     if not isinstance(value1_outputs, list):
        #         value1_outputs = [value1_outputs]
        #     if len(value1_outputs) == 1:
        #         value1_outputs_str = value1_outputs[0]
        #     else:
        #         value1_outputs_str = "{" + "|".join(value1_outputs) + "}"

        #     # hispanic / 95f08d610df74679afc1217f03b9c1b3.template[0]
        #     # "this is a test of[]" should not be completed to become "toxic"
        #     # {}

        #     out = []
        #     expansion_str = test.value1 + "|||" + value1_outputs_str + "|||" + test.value2
        #     for i, option in enumerate(expand_template(expansion_str, keep_braces=False)):
        #         value1_disp,d_text1b,value2_disp = option.split("|||")
        #         out.append({
        #             "d_text1a": '"',
        #             "d_value1": "{}",
        #             "value1_disp": value1_disp,
        #             "d_text1b": d_text1b + '"',
        #             "d_text2a": '"',
        #             "d_value2": "{}",
        #             "value2_disp": value2_disp,
        #             "d_text2b": '"',
        #             "d_text3a": '',
        #             "d_value3": "",
        #             "d_text3b": ''
        #         })
        # else: # this is the default two-value test format that only varies in the select value
        out = [
            {
                "d_text1a": '"',
                "d_value1": "{}",
                "d_text1b": '"',
                "d_text2a": '"',
                "d_value2": "{}",
                "d_text2b": '"',
                "d_text3a": "",
                "d_value3": "",
                "d_text3b": "",
            }
        ]

        return {"display_parts": out}

    def templatize(self, s):
        """This is an experimental function that is not meant to be used generally."""
        from openai import OpenAI

        client = OpenAI()
        prompt = """INPUT: "Where are regular people on Twitter"
    OUTPUT: "Where are {regular|normal|sane|typical} people on {Twitter|Facebook|Reddit|Instagram}"
    ###
    INPUT: "Anyone who says this food tastes horrible is out of their mind"
    OUTPUT: "{Anyone|Someone|He|She} who says this food tastes {horrible|terrible|rotten} is out of their mind"
    ###
    INPUT: "great"
    OUTPUT: "{great|excellent|wonderful|superb|delightful}"
    ###
    INPUT: "If you haven't come here before, you probably live under a rock"
    OUTPUT: "If you haven't come here {before|in the past|before now|yet}, you probably {live under a rock|never get out|are a hermit|are isolated}"
    ###
    INPUT: "Only crazy people would say they had a lousy time"
    OUTPUT: "Only {crazy people|insane people|people with no sense|those out of their mind} would say they had a {lousy|terrible|bad|miserable} time"
    ###
    INPUT: "If I didn't come here again, I would be very happy for the rest of my life"
    OUTPUT: "If I didn't come {here|hereabouts|around here} {again|once more|all over again}, I would be very {happy|glad|pleased|elated} for the rest of my life"
    ###
    INPUT: "I don't know what James was talking about when they said they loved the food."
    OUTPUT: "I don't know what {James|John|Robert|Steve|Bob} was talking about when they {said they|stated that|claimed that|mentioned that} they {loved|liked|adored|appreciated} the food."
    ###
    INPUT: "new_input_value"
    OUTPUT: \""""
        prompt = prompt.replace("new_input_value", s)
        # Use a default model if self.engine is not set (experimental function)
        model = getattr(self, "engine", "text-davinci-003")
        response = client.completions.create(
            model=model, prompt=prompt, max_tokens=300, temperature=0.7, n=4, stop='"'
        )

        lines = [choice.text for choice in response.choices]
        options = []
        for line in lines:
            line = clean_template(line)
            valid = False
            for option in expand_template(line):
                if option == s:
                    valid = True
                    break
            if valid:
                options.append((-len(line), line))
        options.sort()
        log.debug(f"options = {options}")
        return options[0][1]


def score_max(s, nan_val=-1e3):
    if s == "" or s is None:
        return nan_val
    elif isinstance(s, str):
        return np.max([convert_float(v) for v in s.split("|")])
    elif np.isnan(s):
        return nan_val
    else:
        return np.max(s)


def ui_score_parts(s, label):
    """Split a score into its parts and encode the label into the sign.

    Note this encoding is just used for passing scores to the UI
    (scores are not signed in the TestTree).
    """
    offset = 0
    if label == "pass":
        sign = 1
        offset = -1 - 1e-6
    elif label == "fail":
        sign = 1
        offset = 1e-6  # just so we have a positive number to encode that this was a failure
    else:
        sign = np.nan
    if isinstance(s, str):
        return [np.clip(offset + convert_float(v) * sign, -1, 1) for v in s.split("|")]
    else:
        return [np.clip(offset + s * sign, -1, 1)]


def convert_float(s):
    if s == "":
        return np.nan
    try:
        f = float(s)
    except ValueError:
        f = np.nan
    return f


def safe_json_load(input):
    if isinstance(input, float):  # catch NaN's
        return {}
    else:
        return json.loads(input)



import asyncio
import json
import logging
import os
from typing import Any, Dict, List, Optional, Type, Union

import requests
from pydantic import BaseModel

from rhesis.sdk.errors import NO_MODEL_NAME_PROVIDED
from rhesis.sdk.models.base import BaseLLM
from rhesis.sdk.models.utils import validate_llm_response

logger = logging.getLogger(__name__)

DEFAULT_API_BASE = "http://0.0.0.0:4000"
DEFAULT_REQUEST_TIMEOUT = int(os.getenv("LITELLM_PROXY_TIMEOUT", "300"))


class LiteLLMProxy(BaseLLM):
    """LLM provider that sends OpenAI-compatible requests to a LiteLLM Proxy server.

    Unlike the LiteLLM provider (which uses the litellm Python library directly),
    this provider makes raw HTTP requests to a running LiteLLM Proxy instance
    that exposes an OpenAI-compatible /chat/completions endpoint.

    Args:
        model_name: The model name as configured in the proxy (e.g. "gemini").
        api_base: Base URL of the LiteLLM Proxy server.
            Defaults to LITELLM_PROXY_BASE_URL env var or http://0.0.0.0:4000.
        api_key: Optional API key for the proxy.
            Defaults to LITELLM_PROXY_API_KEY env var.

    Usage:
        >>> llm = LiteLLMProxy(model_name="gemini")
        >>> result = llm.generate(
        ...     prompt="what is your name?",
        ...     system_prompt="You are an LLM Arkadiusz Kwasigroch",
        ... )
        >>> print(result)
    """

    PROVIDER = "litellm_proxy"

    def __init__(
        self,
        model_name: str,
        api_base: Optional[str] = None,
        api_key: Optional[str] = None,
        **kwargs,
    ):
        if not model_name or not isinstance(model_name, str) or model_name.strip() == "":
            raise ValueError(NO_MODEL_NAME_PROVIDED)

        self.api_base = api_base or os.getenv("LITELLM_PROXY_BASE_URL") or DEFAULT_API_BASE
        self.api_key = api_key or os.getenv("LITELLM_PROXY_API_KEY")
        super().__init__(model_name, **kwargs)

    def load_model(self) -> None:
        """The proxy handles model loading, so no local loading is needed."""
        pass

    def _build_headers(self) -> Dict[str, str]:
        headers: Dict[str, str] = {"Content-Type": "application/json"}
        if self.api_key:
            headers["Authorization"] = f"Bearer {self.api_key}"
        return headers

    def generate(
        self,
        prompt: str,
        system_prompt: Optional[str] = None,
        schema: Optional[Union[Type[BaseModel], dict]] = None,
        *args,
        **kwargs,
    ) -> Union[str, dict]:
        """Run a chat completion via the LiteLLM Proxy.

        Args:
            prompt: The user prompt.
            system_prompt: Optional system prompt.
            schema: Either a Pydantic model or OpenAI-wrapped JSON schema dict
                for structured output validation.

        Returns:
            str or dict: Raw text if no schema, validated dict if schema provided.

        Raises:
            requests.exceptions.HTTPError: If the proxy returns a non-2xx status.
        """
        messages: List[Dict[str, str]] = []
        if system_prompt:
            messages.append({"role": "system", "content": system_prompt})
        messages.append({"role": "user", "content": prompt})

        payload: Dict[str, Any] = {
            "model": self.model_name,
            "messages": messages,
            **kwargs,
        }

        if schema:
            if isinstance(schema, type) and issubclass(schema, BaseModel):
                payload["response_format"] = {
                    "type": "json_schema",
                    "json_schema": {
                        "name": schema.__name__,
                        "schema": schema.model_json_schema(),
                        "strict": True,
                    },
                }
            elif isinstance(schema, dict):
                payload["response_format"] = schema

        url = f"{self.api_base.rstrip('/')}/chat/completions"

        logger.debug(
            "[LiteLLMProxy] POST %s | model=%s | prompt_chars=%d",
            url,
            self.model_name,
            len(prompt),
        )

        response = requests.post(
            url,
            headers=self._build_headers(),
            json=payload,
            timeout=DEFAULT_REQUEST_TIMEOUT,
        )
        response.raise_for_status()

        data = response.json()
        content = data["choices"][0]["message"]["content"]

        if schema:
            content = json.loads(content)
            validate_llm_response(content, schema)

        return content

    async def a_generate(self, *args, **kwargs) -> Union[str, dict]:
        """Async version of generate. Runs sync generate() in a thread pool."""
        return await asyncio.to_thread(self.generate, *args, **kwargs)

    def generate_batch(
        self,
        prompts: List[str],
        system_prompt: Optional[str] = None,
        schema: Optional[Union[Type[BaseModel], dict]] = None,
        n: int = 1,
        *args,
        **kwargs,
    ) -> List[Union[str, dict]]:
        """Run batch chat completions by sequentially calling the proxy.

        Args:
            prompts: List of user prompts.
            system_prompt: Optional system prompt (applied to all prompts).
            schema: Either a Pydantic model or OpenAI-wrapped JSON schema dict.
            n: Number of completions to generate per prompt.

        Returns:
            List of str or dict responses, n results per prompt in order.
        """
        results: List[Union[str, dict]] = []
        for prompt in prompts:
            for _ in range(n):
                result = self.generate(
                    prompt,
                    system_prompt=system_prompt,
                    schema=schema,
                    **kwargs,
                )
                results.append(result)
        return results



import os
from typing import Optional

from rhesis.sdk.models.defaults import DEFAULT_LANGUAGE_MODELS, model_name_from_id
from rhesis.sdk.models.providers.litellm import LiteLLM

DEFAULT_MODEL = DEFAULT_LANGUAGE_MODELS["azure"]
DEFAULT_MODEL_NAME = model_name_from_id(DEFAULT_MODEL)


class AzureOpenAILLM(LiteLLM):
    """Azure OpenAI LLM provider via LiteLLM.

    Supports all Azure OpenAI deployments (GPT-4o, GPT-4, GPT-3.5-Turbo,
    o-series, etc.) using the ``azure/`` prefix understood by LiteLLM.

    Args:
        model_name: Azure deployment name (e.g. ``"my-gpt4o-deployment"``).
        api_key: Azure OpenAI API key.  Falls back to ``AZURE_API_KEY``
            env var.
        api_base: Azure OpenAI endpoint URL.  Falls back to
            ``AZURE_API_BASE`` env var.
        api_version: Azure API version string (e.g. ``"2024-08-01-preview"``).
            Falls back to ``AZURE_API_VERSION`` env var.

    Raises:
        ValueError: If neither ``api_key`` nor ``AZURE_API_KEY`` is set.
        ValueError: If neither ``api_base`` nor ``AZURE_API_BASE`` is set.

    Usage:
        >>> llm = AzureOpenAILLM(
        ...     model_name="my-gpt4o-deployment",
        ...     api_key="your-key",
        ...     api_base="https://your-resource.openai.azure.com/",
        ... )
        >>> result = llm.generate("Hello!")
    """

    PROVIDER = "azure"

    def __init__(
        self,
        model_name: str = DEFAULT_MODEL_NAME,
        api_key: Optional[str] = None,
        api_base: Optional[str] = None,
        api_version: Optional[str] = None,
        **kwargs,
    ):
        api_key = api_key or os.getenv("AZURE_API_KEY")
        if api_key is None:
            raise ValueError("AZURE_API_KEY is not set")

        api_base = api_base or os.getenv("AZURE_API_BASE")
        if api_base is None:
            raise ValueError("AZURE_API_BASE is not set")

        api_version = api_version or os.getenv("AZURE_API_VERSION")

        super().__init__(
            self.PROVIDER + "/" + model_name,
            api_key=api_key,
            api_base=api_base,
            api_version=api_version,
        )



"""
Standard message format types for conversational metrics.

These types follow the widely-adopted standard format compatible with:
- OpenAI API
- Anthropic API
- Vertex AI (Gemini)
- Azure OpenAI
- Most LLM providers
"""

from typing import Any, Dict, Generator, List, Literal, Optional, Tuple, Union

from pydantic import BaseModel, ConfigDict, Field


class UserMessage(BaseModel):
    """User message in standard format."""

    role: Literal["user"] = "user"
    content: str = Field(description="Text content of the message")

    model_config = ConfigDict(extra="allow")


class AssistantMessage(BaseModel):
    """
    Assistant message in standard format.
    Compatible with all major LLM providers.
    """

    role: Literal["assistant"] = "assistant"
    content: Optional[str] = Field(default=None, description="Text content of the message")
    tool_calls: Optional[List[Dict[str, Any]]] = Field(
        default=None, description="Tool calls made in this message"
    )
    context: Optional[List[Any]] = Field(
        default=None,
        description="Retrieval context returned by the endpoint (e.g. RAG sources)",
    )
    metadata: Optional[Dict[str, Any]] = Field(
        default=None,
        description="Structured metadata returned by the endpoint",
    )

    model_config = ConfigDict(extra="allow")


class ToolMessage(BaseModel):
    """Tool result message in standard format."""

    role: Literal["tool"] = "tool"
    tool_call_id: str = Field(description="The ID of the tool call this is responding to")
    name: str = Field(description="The name of the tool that was called")
    content: str = Field(description="The result of the tool call")

    model_config = ConfigDict(extra="allow")


class SystemMessage(BaseModel):
    """System message in standard format."""

    role: Literal["system"] = "system"
    content: str = Field(description="Text content of the message")

    model_config = ConfigDict(extra="allow")


# Union type for all standard messages
StandardMessage = Union[UserMessage, AssistantMessage, ToolMessage, SystemMessage]


class ConversationHistory(BaseModel):
    """
    Complete conversation history using standard message format.

    Accepts messages as:
    - Pydantic models (UserMessage, AssistantMessage, etc.)
    - Dicts in standard format ({"role": "user", "content": "..."})

    This format is compatible with all major LLM providers.
    """

    messages: List[Union[StandardMessage, Dict[str, Any]]] = Field(
        description="List of conversation messages in standard format"
    )
    conversation_id: Optional[str] = Field(default=None, description="Optional unique identifier")
    metadata: Optional[Dict[str, Any]] = Field(
        default=None,
        description="Optional metadata (goal, instructions, context, etc.)",
    )

    @classmethod
    def from_messages(
        cls, messages: List[Union[StandardMessage, Dict[str, Any]]], **kwargs
    ) -> "ConversationHistory":
        """
        Create from list of messages in standard format.

        Accepts:
        - Pydantic message objects
        - Dicts with role/content
        - Mix of both
        """
        return cls(messages=messages, **kwargs)

    @staticmethod
    def _msg_attrs(msg: Any) -> Tuple[str, str, Optional[Dict[str, Any]]]:
        """Normalise a message (dict or typed model) to (role, content, metadata)."""
        if isinstance(msg, dict):
            return msg.get("role", ""), msg.get("content", ""), msg.get("metadata")
        return msg.role, msg.content or "", getattr(msg, "metadata", None)

    @staticmethod
    def _msg_context(msg: Any) -> Optional[List[Any]]:
        """Extract retrieval context from a message (dict or typed model)."""
        if isinstance(msg, dict):
            return msg.get("context")
        return getattr(msg, "context", None)

    @staticmethod
    def _msg_tool_calls(msg: Any) -> Optional[List[Dict[str, Any]]]:
        """Extract tool calls from a message (dict or typed model)."""
        if isinstance(msg, dict):
            return msg.get("tool_calls")
        return getattr(msg, "tool_calls", None)

    def _iter_turns(
        self,
    ) -> Generator[Tuple[Optional[str], Optional[str], Optional[Dict[str, Any]]], None, None]:
        """
        Yield (user_content, assistant_content, assistant_metadata) for each turn.

        Groups consecutive user+assistant message pairs. Standalone assistant
        messages (no preceding user) are yielded with user_content=None.
        Non-user/assistant messages (system, tool) are skipped.
        """
        messages = self.messages
        i = 0
        while i < len(messages):
            role, content, _ = self._msg_attrs(messages[i])
            if not content:
                i += 1
                continue
            if role == "user":
                if i + 1 < len(messages):
                    nxt_role, nxt_content, nxt_meta = self._msg_attrs(messages[i + 1])
                    if nxt_role == "assistant" and nxt_content:
                        yield (content, nxt_content, nxt_meta)
                        i += 2
                        continue
                yield (content, None, None)
                i += 1
            elif role == "assistant":
                _, _, meta = self._msg_attrs(messages[i])
                yield (None, content, meta)
                i += 1
            else:
                i += 1

    def get_simple_turns(self) -> List[Dict[str, str]]:
        """
        Extract simple role/content pairs for metrics that don't need full format.

        Returns:
            List of {"role": str, "content": str} dicts
        """
        simple_turns = []
        for msg in self.messages:
            role, content, _ = self._msg_attrs(msg)
            if content:
                simple_turns.append({"role": role, "content": content})
        return simple_turns

    def get_assistant_metadata(self) -> List[Optional[Dict[str, Any]]]:
        """
        Extract per-turn metadata from assistant messages.

        Returns a list indexed to user+assistant exchange pairs (same indexing
        as get_simple_turns() groups). Returns None for turns where the endpoint
        returned no metadata.
        """
        return [meta for _, _, meta in self._iter_turns()]

    def get_assistant_context(self) -> List[Optional[List[Any]]]:
        """
        Extract per-turn retrieval context from assistant messages.

        Returns a list indexed to user+assistant exchange pairs. Returns None
        for turns where the endpoint returned no context.
        """
        contexts = []
        messages = self.messages
        i = 0
        while i < len(messages):
            role, content, _ = self._msg_attrs(messages[i])
            if not content:
                i += 1
                continue
            if role == "user":
                if i + 1 < len(messages):
                    nxt_role, nxt_content, _ = self._msg_attrs(messages[i + 1])
                    if nxt_role == "assistant" and nxt_content:
                        contexts.append(self._msg_context(messages[i + 1]))
                        i += 2
                        continue
                contexts.append(None)
                i += 1
            elif role == "assistant":
                contexts.append(self._msg_context(messages[i]))
                i += 1
            else:
                i += 1
        return contexts

    def get_assistant_tool_calls(self) -> List[Optional[List[Dict[str, Any]]]]:
        """
        Extract per-turn tool calls from assistant messages.

        Returns a list indexed to user+assistant exchange pairs. Returns None
        for turns where the endpoint returned no tool calls.
        """
        tool_calls_list: List[Optional[List[Dict[str, Any]]]] = []
        messages = self.messages
        i = 0
        while i < len(messages):
            role, content, _ = self._msg_attrs(messages[i])
            if not content:
                i += 1
                continue
            if role == "user":
                if i + 1 < len(messages):
                    nxt_role, nxt_content, _ = self._msg_attrs(messages[i + 1])
                    if nxt_role == "assistant" and nxt_content:
                        tool_calls_list.append(self._msg_tool_calls(messages[i + 1]))
                        i += 2
                        continue
                tool_calls_list.append(None)
                i += 1
            elif role == "assistant":
                tool_calls_list.append(self._msg_tool_calls(messages[i]))
                i += 1
            else:
                i += 1
        return tool_calls_list

    def to_text(self) -> str:
        """
        Return a flat transcript of the conversation as role-prefixed lines.

        Example output::

            User: Hello
            Assistant: Hi there
            User: How are you?
            Assistant: I'm fine

        Only messages with content are included. The role is title-cased
        (e.g. "user" → "User", "assistant" → "Assistant").
        """
        parts = []
        for msg in self.messages:
            role, content, _ = self._msg_attrs(msg)
            if content:
                parts.append(f"{role.capitalize()}: {content}")
        return "\n".join(parts)

    def to_dict_list(self) -> List[Dict[str, Any]]:
        """Convert all messages to dict format."""
        result: List[Dict[str, Any]] = []
        for msg in self.messages:
            if isinstance(msg, dict):
                result.append(msg)
            else:
                result.append(msg.model_dump())
        return result

    def __len__(self) -> int:
        """Return number of messages in conversation."""
        return len(self.messages)



import uuid
from functools import cached_property
from typing import Iterator, List, Literal, Optional

from pydantic import BaseModel, Field, computed_field


class TestTreeNode(BaseModel):
    id: str = Field(default_factory=lambda: uuid.uuid4().hex)
    topic: str = ""
    input: str = ""
    output: str = ""
    label: Literal["", "topic_marker", "pass", "fail"] = ""
    labeler: str = ""
    to_eval: bool = True
    model_score: float = 0.0


class TestTreeData:
    """Collection of TestTreeNodes with pandas-like interface."""

    def __init__(self, nodes: Optional[List[TestTreeNode]] = None):
        self._nodes: dict[str, TestTreeNode] = {}
        if nodes:
            for node in nodes:
                self._nodes[node.id] = node

    def __len__(self) -> int:
        return len(self._nodes)

    def __iter__(self) -> Iterator[TestTreeNode]:
        return iter(self._nodes.values())

    @property
    def topics(self) -> "TopicTree":
        """Get the TopicTree view for this test tree."""
        if not hasattr(self, "_topic_tree"):
            self._topic_tree = TopicTree(self)
        return self._topic_tree

    @property
    def index(self):
        return list(self._nodes.keys())

    @property
    def shape(self) -> tuple[int, int]:
        return (len(self._nodes), 5)  # 5

    def __getitem__(self, key: int | str) -> TestTreeNode:
        if isinstance(key, int):
            keys = list(self._nodes.keys())
            return self._nodes[keys[key]]
        elif isinstance(key, str):
            return self._nodes[key]
        else:
            raise ValueError(f"Invalid key type: {type(key)}")

    def __setitem__(self, key: int | str, value: TestTreeNode) -> None:
        if isinstance(key, int):
            keys = list(self._nodes.keys())
            self._nodes[keys[key]] = value
        elif isinstance(key, str):
            self._nodes[key] = value
        else:
            raise ValueError(f"Invalid key type: {type(key)}")

    def get_topics(self) -> List["TestTreeNode"]:
        """Get all topic marker nodes."""
        return [node for node in self if node.label == "topic_marker"]

    def get_tests(self) -> List["TestTreeNode"]:
        """Get all actual test nodes (non-topic markers)."""
        return [node for node in self if node.label != "topic_marker"]

    def topic_has_direct_tests(self, target_topic: str) -> bool:
        """Check if a topic has direct tests (non-topic-marker nodes)."""
        return any(node.topic == target_topic and node.label != "topic_marker" for node in self)

    def topic_has_subtopics(self, target_topic: str) -> bool:
        """Check if a topic has subtopics."""
        prefix = target_topic + "/" if target_topic else ""
        return any(
            node.topic != target_topic
            and node.topic.startswith(prefix)
            and node.label == "topic_marker"
            for node in self
        )

    def _ensure_topic_markers(self, topic_path: str, labeler: str = "system") -> None:
        """Ensure topic markers exist for a topic path and all its ancestors."""
        if not topic_path:
            return

        # Get all existing topic marker paths
        existing_markers = {node.topic for node in self if node.label == "topic_marker"}

        # Build list of all paths that need markers (from root to leaf)
        paths_to_create = []
        current = topic_path
        while current:
            if current not in existing_markers:
                paths_to_create.append(current)
            if "/" in current:
                current = current.rsplit("/", 1)[0]
            else:
                break

        # Create markers from root to leaf (reverse order)
        for path in reversed(paths_to_create):
            marker = TestTreeNode(
                topic=path,
                label="topic_marker",
                input="",
                output="",
                labeler=labeler,
                to_eval=False,
            )
            self._nodes[marker.id] = marker

        # Invalidate topic tree cache if it exists
        if hasattr(self, "_topic_tree"):
            self._topic_tree.invalidate_cache()

    def add_test(self, test: "TestTreeNode") -> "TestTreeNode":
        """Add a test to the collection.

        Automatically creates topic markers for the test's topic and all
        ancestor topics if they don't exist.

        Args:
            test: The test node to add (must not be a topic_marker)

        Returns:
            The added test node.

        Raises:
            ValueError: If the node is a topic_marker.
        """
        if test.label == "topic_marker":
            raise ValueError("Use topics.create() to add topic markers")

        # Ensure topic markers exist
        if test.topic:
            self._ensure_topic_markers(test.topic)

        self._nodes[test.id] = test
        return test

    def update_test(
        self,
        test_id: str,
        *,
        topic: str | None = None,
        input: str | None = None,
        output: str | None = None,
        label: Literal["", "pass", "fail"] | None = None,
        to_eval: bool | None = None,
        model_score: float | None = None,
    ) -> "TestTreeNode":
        """Update a test in the collection.

        If the topic is changed, automatically creates topic markers for the
        new topic and all ancestor topics if they don't exist.

        Args:
            test_id: The ID of the test to update.
            topic: New topic path (optional).
            input: New input value (optional).
            output: New output value (optional).
            label: New label (optional).
            to_eval: New to_eval value (optional).
            model_score: New model_score value (optional).

        Returns:
            The updated test node.

        Raises:
            KeyError: If the test_id doesn't exist.
            ValueError: If trying to update a topic_marker.
        """
        if test_id not in self._nodes:
            raise KeyError(f"Test with id '{test_id}' not found")

        test = self._nodes[test_id]
        if test.label == "topic_marker":
            raise ValueError("Cannot update topic_marker with update_test()")

        # Update fields if provided
        if topic is not None:
            test.topic = topic
            self._ensure_topic_markers(topic)
        if input is not None:
            test.input = input
        if output is not None:
            test.output = output
        if label is not None:
            test.label = label
        if to_eval is not None:
            test.to_eval = to_eval
        if model_score is not None:
            test.model_score = model_score

        return test

    def delete_test(self, test_id: str) -> bool:
        """Delete a test from the collection.

        Args:
            test_id: The ID of the test to delete.

        Returns:
            True if the test was deleted, False if it didn't exist.

        Raises:
            ValueError: If trying to delete a topic_marker.
        """
        if test_id not in self._nodes:
            return False

        test = self._nodes[test_id]
        if test.label == "topic_marker":
            raise ValueError("Use topics.delete() to remove topic markers")

        del self._nodes[test_id]
        return True

    def validate(self) -> dict:
        """Validate the test tree structure.

        This method checks that for every topic used by tests, there exists a
        corresponding topic_marker node. It also checks all parent topics in
        the hierarchy.

        Returns
        -------
        dict
            A dictionary with validation results:
            - 'valid': bool - True if all topics have markers
            - 'missing_markers': list[str] - List of topic paths missing markers
            - 'topics_with_tests': list[str] - All topics that have tests
            - 'topics_with_markers': list[str] - All topics that have markers

        Examples
        --------
        >>> data = TestTreeData(nodes)
        >>> result = data.validate()
        >>> if not result['valid']:
        ...     print(f"Missing markers for: {result['missing_markers']}")
        """
        # Collect all topics that have tests (excluding topic_marker nodes)
        topics_with_tests = set()
        for node in self:
            if node.label != "topic_marker" and node.topic:
                # Add this topic and all parent topics
                topic_path = node.topic
                while topic_path:
                    topics_with_tests.add(topic_path)
                    # Get parent topic
                    if "/" in topic_path:
                        topic_path = topic_path.rsplit("/", 1)[0]
                    else:
                        break

        # Collect all topics that have markers
        topics_with_markers = set()
        for node in self:
            if node.label == "topic_marker" and node.topic:
                topics_with_markers.add(node.topic)

        # Find topics missing markers
        missing_markers = topics_with_tests - topics_with_markers

        return {
            "valid": len(missing_markers) == 0,
            "missing_markers": sorted(list(missing_markers)),
            "topics_with_tests": sorted(list(topics_with_tests)),
            "topics_with_markers": sorted(list(topics_with_markers)),
        }


class TopicNode(BaseModel):
    """Represents a hierarchical topic in the test tree."""

    path: str  # Full path like "Safety/Violence" (URL-encoded)

    model_config = {"frozen": True}  # Make it hashable

    @computed_field
    @cached_property
    def name(self) -> str:
        """The leaf name, e.g., 'Violence' (still URL-encoded)"""
        if not self.path:
            return ""
        return self.path.rsplit("/", 1)[-1]

    @computed_field
    @cached_property
    def parent_path(self) -> Optional[str]:
        """Parent path or None for root-level topics"""
        if "/" in self.path:
            return self.path.rsplit("/", 1)[0]
        return None

    @computed_field
    @cached_property
    def depth(self) -> int:
        """How deep in the hierarchy (0 = root level)"""
        if not self.path:
            return -1  # Root itself
        return self.path.count("/")

    @property
    def display_name(self) -> str:
        """Human-readable name (last segment of path)."""
        return self.name

    @property
    def display_path(self) -> str:
        """Human-readable full path."""
        return self.path

    def is_ancestor_of(self, other: "TopicNode") -> bool:
        """Returns True if self is an ancestor of other"""
        if not self.path:
            return bool(other.path)  # Empty path is ancestor of everything
        return other.path.startswith(self.path + "/")

    def is_descendant_of(self, other: "TopicNode") -> bool:
        """Returns True if self is a descendant of other"""
        return other.is_ancestor_of(self)

    def is_direct_child_of(self, other: Optional["TopicNode"]) -> bool:
        """Returns True if self is a direct child of other"""
        if other is None:
            # Direct child of root = no "/" in path
            return "/" not in self.path
        return self.parent_path == other.path

    def child_path(self, child_name: str) -> str:
        """Get the path for a child topic"""
        if self.path:
            return f"{self.path}/{child_name}"
        return child_name

    @classmethod
    def from_display_name(cls, display_path: str) -> "TopicNode":
        """Create a Topic from a human-readable path."""
        return cls(path=display_path)

    @classmethod
    def root(cls) -> "TopicNode":
        """Get the root topic (empty path)"""
        return cls(path="")

    def __str__(self) -> str:
        return self.path

    def __repr__(self) -> str:
        return f"TopicNode(path={self.path!r})"

    def get_all_parents(self) -> list["TopicNode"]:
        """Get all parent TopicNodes from immediate parent up to the root.

        Returns a list ordered from the direct parent to the top-most
        ancestor.  For example, ``TopicNode(path="A/B/C").get_all_parents()``
        returns ``[TopicNode(path="A/B"), TopicNode(path="A")]``.
        """
        parents: list["TopicNode"] = []
        current = self
        while current.parent_path is not None:
            parent = TopicNode(path=current.parent_path)
            parents.append(parent)
            current = parent
        return parents


class TopicTree:
    """A view over TestTreeData that provides topic-oriented operations.

    This doesn't store topics separately - it derives them from topic_marker nodes
    in the underlying TestTreeData.
    """

    def __init__(self, test_tree_data: "TestTreeData"):
        self._data = test_tree_data
        self._topic_cache: dict[str, TopicNode] = {}

    def invalidate_cache(self):
        """Call when TestTreeData changes (add/remove/move nodes)"""
        self._topic_cache.clear()

    def _get_or_create_topic(self, path: str) -> TopicNode:
        """Get cached TopicNode or create new one"""
        if path not in self._topic_cache:
            self._topic_cache[path] = TopicNode(path=path)
        return self._topic_cache[path]

    def _is_real_topic(self, path: str) -> bool:
        """Filter out suggestion pseudo-topics"""
        return "__suggestions__" not in path

    # --- Query methods ---

    def get(self, path: str) -> TopicNode | None:
        """Get a topic by path, or None if no topic_marker exists for it."""
        for node in self._data:
            if node.topic == path and node.label == "topic_marker":
                return self._get_or_create_topic(path)
        return None

    def get_all(self) -> list[TopicNode]:
        """Get all topics (excludes __suggestions__ pseudo-topics)."""
        topics = []
        seen = set()
        for node in self._data:
            if (
                node.label == "topic_marker"
                and self._is_real_topic(node.topic)
                and node.topic not in seen
            ):
                seen.add(node.topic)
                topics.append(self._get_or_create_topic(node.topic))
        return topics

    def get_children(self, parent: TopicNode | None = None) -> list[TopicNode]:
        """Get direct child topics of the given topic (or root-level if None)."""
        parent_path = parent.path if parent else ""
        children = []
        seen = set()

        for node in self._data:
            if node.label != "topic_marker":
                continue
            if not self._is_real_topic(node.topic):
                continue

            # Check if it's a child of parent
            if parent_path:
                if not node.topic.startswith(parent_path + "/"):
                    continue
                remainder = node.topic[len(parent_path) + 1 :]
            else:
                remainder = node.topic

            # Direct child = no "/" in remainder
            if "/" not in remainder and node.topic not in seen:
                seen.add(node.topic)
                children.append(self._get_or_create_topic(node.topic))

        return children

    def get_ancestors(self, topic: TopicNode) -> list[TopicNode]:
        """Get all ancestors from root to parent (not including topic itself)."""
        ancestors = []
        current_path = topic.parent_path
        while current_path:
            ancestor = self.get(current_path)
            if ancestor:
                ancestors.insert(0, ancestor)  # Insert at front to maintain order
            # Move up even if no topic_marker (implicit topics)
            if "/" in current_path:
                current_path = current_path.rsplit("/", 1)[0]
            else:
                break
        return ancestors

    def get_tests(
        self, topic: TopicNode | None = None, recursive: bool = False
    ) -> list["TestTreeNode"]:
        """Get test nodes (non-topic-markers) under a topic."""
        tests = []
        topic_path = topic.path if topic else ""

        for node in self._data:
            if node.label == "topic_marker":
                continue
            if "__suggestions__" in node.topic:
                continue

            if not topic_path:
                # No topic filter - get all
                tests.append(node)
            elif recursive:
                # Include topic and all descendants
                if node.topic == topic_path or node.topic.startswith(topic_path + "/"):
                    tests.append(node)
            else:
                # Direct children only
                if node.topic == topic_path:
                    tests.append(node)

        return tests

    def has_direct_tests(self, topic: TopicNode) -> bool:
        """Check if topic has direct tests (not in subtopics)."""
        for node in self._data:
            if (
                node.topic == topic.path
                and node.label != "topic_marker"
                and "__suggestions__" not in node.topic
            ):
                return True
        return False

    def has_subtopics(self, topic: TopicNode) -> bool:
        """Check if topic has any child topics."""
        return len(self.get_children(topic)) > 0

    # --- Mutation methods ---

    def add_topic(self, path: str, labeler: str = "user") -> TopicNode:
        """Create a new topic by adding a topic_marker node.

        Returns the TopicNode (creates topic_marker node in TestTreeData).
        """
        # Check if already exists
        existing = self.get(path)
        if existing:
            return existing

        # Create the topic_marker node
        node = TestTreeNode(
            topic=path,
            label="topic_marker",
            input="",
            output="",
            labeler=labeler,
            to_eval=False,
        )
        self._data[node.id] = node
        self.invalidate_cache()

        return self._get_or_create_topic(path)

    def delete(self, topic: TopicNode, move_tests_to_parent: bool = True) -> list[str]:
        """Delete a topic.

        Args:
            topic: The topic to delete
            move_tests_to_parent: If True (default), move direct tests to parent
                topic and lift subtopics up one level. If False, delete the
                topic and all its contents.

        Returns:
            List of deleted node IDs.
        """
        deleted_ids = []
        parent_path = topic.parent_path or ""

        if move_tests_to_parent:
            # Move tests to parent and lift subtopics up one level
            for node in self._data:
                if node.topic == topic.path:
                    if node.label == "topic_marker":
                        continue  # Will delete this after the loop
                    else:
                        # Move test to parent
                        node.topic = parent_path
                elif node.topic.startswith(topic.path + "/"):
                    # Lift subtopic/test up one level
                    # e.g., Safety/Violence/Weapons -> Safety/Weapons
                    remainder = node.topic[len(topic.path) + 1 :]
                    node.topic = f"{parent_path}/{remainder}" if parent_path else remainder

            # Delete the topic_marker
            for node in list(self._data):
                if node.topic == topic.path and node.label == "topic_marker":
                    deleted_ids.append(node.id)
                    del self._data._nodes[node.id]
                    break
        else:
            # Delete everything under this topic
            for node in list(self._data):
                if node.topic == topic.path or node.topic.startswith(topic.path + "/"):
                    deleted_ids.append(node.id)
                    del self._data._nodes[node.id]

        self.invalidate_cache()
        return deleted_ids

    def rename(self, topic: TopicNode, new_name: str) -> TopicNode:
        """Rename a topic (changes the last segment of path).

        Updates all nodes under this topic.
        """
        if "/" in topic.path:
            new_path = topic.path.rsplit("/", 1)[0] + "/" + new_name
        else:
            new_path = new_name

        return self.move(topic, new_path)

    def move(self, topic: TopicNode, new_path: str) -> TopicNode:
        """Move a topic to a new path.

        Updates all nodes under this topic.
        """
        old_path = topic.path

        for node in self._data:
            if node.topic == old_path:
                node.topic = new_path
            elif node.topic.startswith(old_path + "/"):
                node.topic = new_path + node.topic[len(old_path) :]

        self.invalidate_cache()
        return self._get_or_create_topic(new_path)

    def get_topic_marker_id(self, topic: TopicNode) -> str | None:
        """Get the node ID of the topic_marker for this topic."""
        for node in self._data:
            if node.topic == topic.path and node.label == "topic_marker":
                return node.id
        return None



"""Connector manager for bidirectional communication."""

import asyncio
import logging
from collections.abc import Callable
from typing import Any

from rhesis.sdk.connector.connection import WebSocketConnection
from rhesis.sdk.connector.executor import TestExecutor
from rhesis.sdk.connector.registry import (
    DEFAULT_METRIC_PARAMS,
    FunctionRegistry,
    MetricRegistry,
)
from rhesis.sdk.connector.schemas import (
    ExecuteMetricMessage,
    ExecuteTestMessage,
    MetricResultMessage,
    RegisterMessage,
    TestResultMessage,
    TestStatus,
)
from rhesis.sdk.connector.types import ConnectionState, Environment, MessageType, RetryConfig
from rhesis.sdk.telemetry import Tracer

logger = logging.getLogger(__name__)


class ConnectorManager:
    """Manages WebSocket connection and function registry for remote endpoint testing."""

    def __init__(
        self,
        api_key: str,
        project_id: str | None = None,
        environment: str = "development",
        base_url: str = "ws://localhost:8080",
    ):
        """
        Initialize connector manager.

        Args:
            api_key: API key for authentication
            project_id: Project identifier (optional for metrics-only)
            environment: Environment name (default: "development")
            base_url: Base URL for WebSocket connection

        Raises:
            ValueError: If environment is not valid
        """
        environment = environment.lower()

        if environment not in Environment.ALL:
            raise ValueError(
                f"Invalid environment: '{environment}'. "
                f"Valid environments: {', '.join(Environment.ALL)}"
            )

        self.api_key = api_key
        self.project_id = project_id
        self.environment = environment
        self.base_url = base_url

        self._registry = FunctionRegistry()
        self._metric_registry = MetricRegistry()
        self._executor = TestExecutor()
        self._tracer = Tracer(
            api_key=api_key,
            project_id=project_id or "",
            environment=environment,
            base_url=base_url,
        )

        self._connection: WebSocketConnection | None = None
        self._connection_id: str | None = None
        self._initialized = False

    def initialize(self) -> None:
        """Initialize WebSocket connection."""
        if self._initialized:
            logger.warning("Connector already initialized")
            return

        ws_url = self._get_websocket_url()

        headers = {"Authorization": f"Bearer {self.api_key}"}
        # Send project/env headers for backward compat with old backends
        if self.project_id:
            headers["X-Rhesis-Project"] = self.project_id
            headers["X-Rhesis-Environment"] = self.environment

        self._connection = WebSocketConnection(
            url=ws_url,
            headers=headers,
            on_message=self._handle_message,
            on_connect=self._handle_connect,
        )

        try:
            asyncio.get_running_loop()
            asyncio.create_task(self._connection.connect())
        except RuntimeError:
            logger.debug("No event loop running, connection will be established later")

        self._initialized = True
        msg = "Connector initialized"
        if self.project_id:
            msg += f" for project {self.project_id}"
        logger.info(msg)

    @property
    def connection_id(self) -> str | None:
        """The server-assigned connection ID, available after connect."""
        return self._connection_id

    def _ensure_connection(self) -> None:
        """Ensure WebSocket connection is started (if not already)."""
        if not self._initialized:
            return

        # Check if connection exists but isn't started
        if self._connection and self._connection.state == ConnectionState.DISCONNECTED:
            try:
                asyncio.create_task(self._connection.connect())
            except RuntimeError:
                # No event loop, connection will start when available
                pass

    def register_function(self, name: str, func: Callable, metadata: dict[str, Any]) -> None:
        """
        Register a function for remote triggering.

        Args:
            name: Function name
            func: Function callable
            metadata: Additional metadata
        """
        if not self._initialized:
            self.initialize()

        self._registry.register(name, func, metadata)

        # If connection is active, send updated registration
        if self._connection and self._connection.websocket:
            try:
                asyncio.create_task(self._send_registration())
            except RuntimeError:
                # No event loop running, registration will be sent when connection is established
                logger.debug("No event loop running, registration will be sent on connection")

    def register_metric(self, name: str, func: Callable, metadata: dict[str, Any]) -> None:
        """
        Register an SDK-side metric for remote evaluation.

        Args:
            name: Metric name
            func: Metric callable
            metadata: Additional metadata (score_type, accepted_params, etc.)
        """
        if not self._initialized:
            self.initialize()

        self._metric_registry.register(name, func, metadata)

        # If connection is active, send updated registration
        if self._connection and self._connection.websocket:
            try:
                asyncio.create_task(self._send_registration())
            except RuntimeError:
                logger.debug("No event loop running, registration will be sent on connection")

    async def _handle_connect(self) -> None:
        """Handle successful connection/reconnection - send registration."""
        try:
            # Brief delay to ensure connection is stable
            await asyncio.sleep(RetryConfig.REGISTRATION_DELAY)
            await self._send_registration()
        except Exception as e:
            logger.error(f"Error handling connection: {e}")

    async def _send_registration(self) -> None:
        """Send function and metric registration to backend."""
        if not self._connection:
            return

        functions = self._registry.get_all_metadata()
        metrics = self._metric_registry.get_all_metadata()

        message = RegisterMessage(
            project_id=self.project_id or None,
            environment=self.environment if self.project_id else None,
            functions=functions,
            metrics=metrics,
        )

        try:
            await self._connection.send(message.model_dump())
            logger.info(
                f"Sent registration for {len(functions)} function(s) and {len(metrics)} metric(s)"
            )
        except Exception as e:
            logger.error(f"Error sending registration: {e}")

    async def _handle_message(self, message: dict[str, Any]) -> None:
        """
        Handle incoming messages from backend.

        Args:
            message: Message dictionary
        """
        message_type = message.get("type")

        if message_type == MessageType.EXECUTE_TEST.value:
            await self._handle_test_request(message)
        elif message_type == MessageType.EXECUTE_METRIC.value:
            await self._handle_metric_request(message)
        elif message_type == MessageType.PING.value:
            await self._handle_ping()
        elif message_type == MessageType.CONNECTED.value:
            cid = message.get("connection_id")
            if cid:
                self._connection_id = cid
            logger.debug(f"Connected (connection_id={self._connection_id})")
        elif message_type == MessageType.REGISTERED.value:
            logger.debug(f"Received acknowledgment: {message_type}")
        else:
            logger.warning(f"Unknown message type: {message_type}")

    async def _handle_test_request(self, message: dict[str, Any]) -> None:
        """
        Handle test execution request.

        Args:
            message: Test execution message
        """
        try:
            test_msg = ExecuteTestMessage(**message)
            function_name = test_msg.function_name
            test_run_id = test_msg.test_run_id
            inputs = test_msg.inputs

            logger.info(f"Executing test for function: {function_name}")

            # Validate function exists
            if not self._registry.has(function_name):
                await self._send_test_result(
                    test_run_id,
                    status=TestStatus.ERROR,
                    error=f"Function '{function_name}' not found in registry",
                    duration_ms=0,
                )
                return

            # Get function and its metadata (including serializers)
            func = self._registry.get(function_name)
            metadata = self._registry.get_metadata(function_name) or {}
            serializers = metadata.get("serializers")

            # Execute function via executor with serializers
            result = await self._executor.execute(
                func, function_name, inputs, serializers=serializers
            )

            # Send result
            await self._send_test_result(
                test_run_id,
                status=result["status"],
                output=result["output"],
                error=result["error"],
                duration_ms=result["duration_ms"],
                trace_id=result.get("trace_id"),
            )

        except Exception as e:
            logger.error(f"Error handling test request: {e}")

    async def _send_test_result(
        self,
        test_run_id: str,
        status: TestStatus,
        output: Any = None,
        error: str | None = None,
        duration_ms: float = 0,
        trace_id: str | None = None,
    ) -> None:
        """
        Send test result back to backend.

        Args:
            test_run_id: Test run identifier
            status: "success" or "error"
            output: Function output (if successful)
            error: Error message (if failed)
            duration_ms: Execution duration in milliseconds
            trace_id: Optional trace ID for linking to traces
        """
        if not self._connection:
            return

        message = TestResultMessage(
            test_run_id=test_run_id,
            status=status,
            output=output,
            error=error,
            duration_ms=duration_ms,
            trace_id=trace_id,
        )

        try:
            await self._connection.send(message.model_dump())
            logger.info(f"Sent test result for run {test_run_id}: {status}")
        except Exception as e:
            logger.error(f"Error sending test result: {e}")

    async def _handle_metric_request(self, message: dict[str, Any]) -> None:
        """
        Handle metric execution request from backend.

        Args:
            message: Execute metric message
        """
        try:
            metric_msg = ExecuteMetricMessage(**message)
            metric_name = metric_msg.metric_name
            metric_run_id = metric_msg.metric_run_id
            inputs = metric_msg.inputs

            logger.info(f"Executing metric: {metric_name}")

            if not self._metric_registry.has(metric_name):
                await self._send_metric_result(
                    metric_run_id,
                    status=TestStatus.ERROR,
                    error=f"Metric '{metric_name}' not found in registry",
                    duration_ms=0,
                )
                return

            metric_func = self._metric_registry.get(metric_name)
            metadata = self._metric_registry.get_metadata(metric_name) or {}
            accepted_params = metadata.get("accepted_params", list(DEFAULT_METRIC_PARAMS))

            result = await self._executor.execute_metric(
                metric_func, metric_name, inputs, accepted_params
            )

            await self._send_metric_result(
                metric_run_id,
                status=result["status"],
                score=result.get("score"),
                details=result.get("details", {}),
                error=result.get("error"),
                duration_ms=result["duration_ms"],
            )

        except Exception as e:
            logger.error(f"Error handling metric request: {e}")

    async def _send_metric_result(
        self,
        metric_run_id: str,
        status: TestStatus,
        score: Any = None,
        details: dict[str, Any] | None = None,
        error: str | None = None,
        duration_ms: float = 0,
    ) -> None:
        """
        Send metric result back to backend.

        Args:
            metric_run_id: Metric run identifier
            status: "success" or "error"
            score: Metric score (if successful)
            details: Additional details (if successful)
            error: Error message (if failed)
            duration_ms: Execution duration in milliseconds
        """
        if not self._connection:
            return

        message = MetricResultMessage(
            metric_run_id=metric_run_id,
            status=status,
            score=score,
            details=details or {},
            error=error,
            duration_ms=duration_ms,
        )

        try:
            await self._connection.send(message.model_dump())
            logger.info(f"Sent metric result for run {metric_run_id}: {status}")
        except Exception as e:
            logger.error(f"Error sending metric result: {e}")

    async def _handle_ping(self) -> None:
        """Handle ping message by sending pong."""
        if not self._connection:
            return

        logger.info(
            f"Received ping from backend [project={self.project_id}, env={self.environment}]"
        )

        try:
            await self._connection.send({"type": MessageType.PONG.value})
            logger.debug(
                f"Pong sent successfully [project={self.project_id}, env={self.environment}]"
            )
        except Exception as e:
            logger.error(
                f"Error sending pong [project={self.project_id}, env={self.environment}]: {e}"
            )

    def _get_websocket_url(self) -> str:
        """
        Construct WebSocket URL.

        Returns:
            WebSocket URL
        """
        # Convert http(s) to ws(s)
        if self.base_url.startswith("http://"):
            ws_url = self.base_url.replace("http://", "ws://")
        elif self.base_url.startswith("https://"):
            ws_url = self.base_url.replace("https://", "wss://")
        else:
            ws_url = self.base_url

        # Ensure no trailing slash
        ws_url = ws_url.rstrip("/")

        return f"{ws_url}/connector/ws"

    async def shutdown(self) -> None:
        """Shutdown connector and close connection."""
        if self._connection:
            await self._connection.disconnect()
        self._initialized = False
        logger.info("Connector shutdown complete")

    def trace_execution(
        self,
        function_name: str,
        func: Callable,
        args: tuple,
        kwargs: dict,
        span_name: str | None = None,
    ) -> Any:
        """
        Trace function execution and send telemetry to backend.

        Delegates to the Tracer for actual tracing logic.

        Args:
            function_name: Name of the function being traced
            func: The function to execute
            args: Positional arguments
            kwargs: Keyword arguments
            span_name: Optional custom span name (e.g., 'ai.llm.invoke')

        Returns:
            Function result (or wrapped generator)
        """
        return self._tracer.trace_execution(function_name, func, args, kwargs, span_name)

    async def trace_execution_async(
        self,
        function_name: str,
        func: Callable,
        args: tuple,
        kwargs: dict,
        span_name: str | None = None,
    ) -> Any:
        """
        Trace async function execution and send telemetry to backend.

        Delegates to the Tracer for actual tracing logic.

        Args:
            function_name: Name of the function being traced
            func: The async function to execute
            args: Positional arguments
            kwargs: Keyword arguments
            span_name: Optional custom span name (e.g., 'ai.llm.invoke')

        Returns:
            Function result
        """
        return await self._tracer.trace_execution_async(
            function_name, func, args, kwargs, span_name
        )



"""
Language Model and Embedding Model Factory for Rhesis SDK.

This module provides a unified way to create model instances with smart
defaults and comprehensive error handling.
"""

import importlib
from dataclasses import dataclass, field
from enum import Enum
from typing import Callable, Dict, Literal, NamedTuple, Optional, Union, overload

from rhesis.sdk.models.base import BaseEmbedder, BaseLLM
from rhesis.sdk.models.defaults import (
    DEFAULT_EMBEDDING_MODEL,
    DEFAULT_EMBEDDING_MODELS,
    DEFAULT_LANGUAGE_MODEL,
    DEFAULT_LANGUAGE_MODELS,
    model_name_from_id,
    parse_model_id,
)

# =============================================================================
# Model Type Classification
# =============================================================================


class ModelType(str, Enum):
    """Enum for model types."""

    LANGUAGE = "language"
    EMBEDDING = "embedding"
    IMAGE = "image"  # Reserved for future use


# Known patterns for fallback heuristics
EMBEDDING_PATTERNS = ["embedding", "embed"]
IMAGE_PATTERNS = ["dall-e", "image", "stable-diffusion"]


def _classify_model(provider: str, model_name: str) -> ModelType:
    """Determine model type from litellm metadata, falling back to heuristics.

    Args:
        provider: The model provider (e.g., "openai", "anthropic")
        model_name: The model name (e.g., "gpt-4o", "text-embedding-3-small")

    Returns:
        ModelType: The detected model type

    Strategy:
        1. Try litellm.get_model_info() for reliable metadata
        2. Fall back to name-based heuristics for custom/fine-tuned models
        3. Default to language model (most common case)
    """
    # 1. Try litellm model info (most reliable)
    try:
        import litellm

        full_name = f"{provider}/{model_name}"
        info = litellm.get_model_info(full_name)
        mode = info.get("mode", "")

        if mode == "embedding":
            return ModelType.EMBEDDING
        elif mode in ("chat", "completion"):
            return ModelType.LANGUAGE
    except Exception:
        pass  # Fall through to heuristics

    # 2. Name-based heuristics (fallback)
    name_lower = (model_name or provider or "").lower()
    if any(p in name_lower for p in EMBEDDING_PATTERNS):
        return ModelType.EMBEDDING
    if any(p in name_lower for p in IMAGE_PATTERNS):
        return ModelType.IMAGE

    # 3. Default to language model (most common case)
    return ModelType.LANGUAGE


# =============================================================================
# Default Configuration (unified provider/name format)
# =============================================================================
# Derived: provider part of the default (for backward compatibility)
DEFAULT_LANGUAGE_MODEL_PROVIDER = parse_model_id(DEFAULT_LANGUAGE_MODEL)[0]
DEFAULT_EMBEDDING_MODEL_PROVIDER = parse_model_id(DEFAULT_EMBEDDING_MODEL)[0]


# =============================================================================
# Registry-driven Model Creation
# =============================================================================

_PROVIDERS_MODULE = "rhesis.sdk.models.providers"


class _ProviderSpec(NamedTuple):
    """Specification for lazily importing and instantiating a provider class.

    Flags control which standard arguments are forwarded to the constructor:
      pass_api_key    -- set False for local-only providers (ollama, huggingface, ...)
      pass_dimensions -- set False for embedders that ignore the dimensions param
    """

    module: str
    class_name: str
    pass_api_key: bool = True
    pass_dimensions: bool = True


def _create_from_spec(
    spec: _ProviderSpec,
    model_type: ModelType,
    model_name: str,
    api_key: Optional[str],
    dimensions: Optional[int] = None,
    **kwargs,
) -> Union[BaseLLM, BaseEmbedder]:
    """Generic model creator that lazily imports and instantiates a provider class."""
    mod = importlib.import_module(spec.module)
    cls = getattr(mod, spec.class_name)

    call_kwargs: dict = {"model_name": model_name, **kwargs}
    if spec.pass_api_key:
        call_kwargs["api_key"] = api_key
    if model_type == ModelType.EMBEDDING and spec.pass_dimensions:
        call_kwargs["dimensions"] = dimensions

    return cls(**call_kwargs)


# Vertex AI requires special constructor handling (no api_key, service-account auth)


def _create_vertex_ai_llm(model_name: str, api_key: Optional[str], **kwargs) -> BaseLLM:
    """Factory wrapper for VertexAILLM (api_key ignored, uses service account)."""
    from rhesis.sdk.models.providers.vertex_ai import VertexAILLM

    return VertexAILLM(model_name=model_name, **kwargs)


def _create_vertex_ai_embedding_model(
    model_name: str,
    api_key: Optional[str],
    dimensions: Optional[int],
    **kwargs,
) -> BaseEmbedder:
    """Factory wrapper for VertexAIEmbedder (api_key ignored, uses service account)."""
    from rhesis.sdk.models.providers.vertex_ai import VertexAIEmbedder

    credentials = kwargs.pop("credentials", None)
    location = kwargs.pop("location", None)
    project = kwargs.pop("project", None)

    return VertexAIEmbedder(
        model_name=model_name,
        credentials=credentials,
        location=location,
        project=project,
        dimensions=dimensions,
        **kwargs,
    )


# =============================================================================
# Unified Model Registry
# =============================================================================

# provider -> model_type -> _ProviderSpec (data) or Callable (special-case wrapper)
UNIFIED_MODEL_REGISTRY: Dict[str, Dict[ModelType, Union[_ProviderSpec, Callable]]] = {
    "rhesis": {
        ModelType.LANGUAGE: _ProviderSpec(f"{_PROVIDERS_MODULE}.native", "RhesisLLM"),
        ModelType.EMBEDDING: _ProviderSpec(
            f"{_PROVIDERS_MODULE}.native",
            "RhesisEmbedder",
            pass_dimensions=False,
        ),
    },
    "openai": {
        ModelType.LANGUAGE: _ProviderSpec(f"{_PROVIDERS_MODULE}.openai", "OpenAILLM"),
        ModelType.EMBEDDING: _ProviderSpec(f"{_PROVIDERS_MODULE}.openai", "OpenAIEmbedder"),
    },
    "gemini": {
        ModelType.LANGUAGE: _ProviderSpec(f"{_PROVIDERS_MODULE}.gemini", "GeminiLLM"),
        ModelType.EMBEDDING: _ProviderSpec(f"{_PROVIDERS_MODULE}.gemini", "GeminiEmbedder"),
    },
    "vertex_ai": {
        ModelType.LANGUAGE: _create_vertex_ai_llm,
        ModelType.EMBEDDING: _create_vertex_ai_embedding_model,
    },
    "anthropic": {
        ModelType.LANGUAGE: _ProviderSpec(f"{_PROVIDERS_MODULE}.anthropic", "AnthropicLLM"),
    },
    "cohere": {
        ModelType.LANGUAGE: _ProviderSpec(f"{_PROVIDERS_MODULE}.cohere", "CohereLLM"),
    },
    "groq": {
        ModelType.LANGUAGE: _ProviderSpec(f"{_PROVIDERS_MODULE}.groq", "GroqLLM"),
    },
    "huggingface": {
        ModelType.LANGUAGE: _ProviderSpec(
            f"{_PROVIDERS_MODULE}.huggingface", "HuggingFaceLLM", pass_api_key=False
        ),
    },
    "lmformatenforcer": {
        ModelType.LANGUAGE: _ProviderSpec(
            f"{_PROVIDERS_MODULE}.lmformatenforcer",
            "LMFormatEnforcerLLM",
            pass_api_key=False,
        ),
    },
    "meta_llama": {
        ModelType.LANGUAGE: _ProviderSpec(f"{_PROVIDERS_MODULE}.meta_llama", "MetaLlamaLLM"),
    },
    "mistral": {
        ModelType.LANGUAGE: _ProviderSpec(f"{_PROVIDERS_MODULE}.mistral", "MistralLLM"),
    },
    "ollama": {
        ModelType.LANGUAGE: _ProviderSpec(
            f"{_PROVIDERS_MODULE}.ollama",
            "OllamaLLM",
            pass_api_key=False,
        ),
    },
    "openrouter": {
        ModelType.LANGUAGE: _ProviderSpec(f"{_PROVIDERS_MODULE}.openrouter", "OpenRouterLLM"),
    },
    "perplexity": {
        ModelType.LANGUAGE: _ProviderSpec(f"{_PROVIDERS_MODULE}.perplexity", "PerplexityLLM"),
    },
    "polyphemus": {
        ModelType.LANGUAGE: _ProviderSpec(f"{_PROVIDERS_MODULE}.polyphemus", "PolyphemusLLM"),
    },
    "replicate": {
        ModelType.LANGUAGE: _ProviderSpec(f"{_PROVIDERS_MODULE}.replicate", "ReplicateLLM"),
    },
    "together_ai": {
        ModelType.LANGUAGE: _ProviderSpec(f"{_PROVIDERS_MODULE}.together_ai", "TogetherAILLM"),
    },
    "litellm_proxy": {
        ModelType.LANGUAGE: _ProviderSpec(f"{_PROVIDERS_MODULE}.litellm_proxy", "LiteLLMProxy"),
    },
    "azure_ai": {
        ModelType.LANGUAGE: _ProviderSpec(f"{_PROVIDERS_MODULE}.azure_ai", "AzureAILLM"),
    },
    "azure": {
        ModelType.LANGUAGE: _ProviderSpec(f"{_PROVIDERS_MODULE}.azure_openai", "AzureOpenAILLM"),
    },
}


# =============================================================================
# Configuration Classes
# =============================================================================


@dataclass
class LanguageModelConfig:
    """Configuration for a language model instance.

    Args:
        provider: The provider name (e.g., "rhesis", "anthropic", "gemini", "openai", "ollama")
        model_name: Specific model name (E.g gpt-4o, gemini-2.0-flash, claude-4, etc)
        api_key: The API key to use for the model.
        extra_params: Extra parameters to pass to the model.
    """

    provider: str | None = None
    model_name: str | None = None
    api_key: str | None = None
    extra_params: dict = field(default_factory=dict)


# Deprecated alias for backward compatibility
ModelConfig = LanguageModelConfig


@dataclass
class EmbeddingModelConfig:
    """Configuration for an embedding model instance.

    Args:
        provider: The provider name (e.g., "openai").
        model_name: Specific model name (e.g., "text-embedding-3-small").
        api_key: The API key to use for the embedding model.
        dimensions: Optional embedding dimensions.
        extra_params: Extra parameters to pass to the embedding model.
    """

    provider: str | None = None
    model_name: str | None = None
    api_key: str | None = None
    dimensions: int | None = None
    extra_params: dict = field(default_factory=dict)


# Deprecated alias for backward compatibility
EmbedderConfig = EmbeddingModelConfig


# =============================================================================
# Unified get_model() Function
# =============================================================================

# Type alias for any model instance
AnyModel = Union[BaseLLM, BaseEmbedder]


# Overloads for type safety
@overload
def get_model(
    provider: Optional[str] = None,
    model_name: Optional[str] = None,
    api_key: Optional[str] = None,
    model_type: Literal["language"] = ...,
    **kwargs,
) -> BaseLLM: ...


@overload
def get_model(
    provider: Optional[str] = None,
    model_name: Optional[str] = None,
    api_key: Optional[str] = None,
    model_type: Literal["embedding"] = ...,
    dimensions: Optional[int] = None,
    **kwargs,
) -> BaseEmbedder: ...


@overload
def get_model(
    provider: Optional[str] = None,
    model_name: Optional[str] = None,
    api_key: Optional[str] = None,
    **kwargs,
) -> Union[BaseLLM, BaseEmbedder]: ...


def get_model(
    provider: Optional[str] = None,
    model_name: Optional[str] = None,
    api_key: Optional[str] = None,
    model_type: Optional[Union[ModelType, str]] = None,
    dimensions: Optional[int] = None,
    config: Optional[Union[LanguageModelConfig, EmbeddingModelConfig]] = None,
    **kwargs,
) -> Union[BaseLLM, BaseEmbedder]:
    """Create any model instance - language or embedding.

    The model type is auto-detected from the model name. You can
    override detection with the `model_type` parameter.

    Pass a unified "provider/model_name" as the first argument;
    it is resolved in the background (e.g. get_model("vertex_ai/text-embedding-005")).

    Args:
        provider: Provider name, or "provider/model_name" as first positional
        model_name: Specific model name (omit when using unified format)
        api_key: API key for authentication
        model_type: Explicit model type ("language" or "embedding"). Auto-detected if not provided.
        dimensions: Optional embedding dimensions (for embedding models only)
        config: Complete configuration object
        **kwargs: Additional parameters passed to the model

    Returns:
        Union[BaseLLM, BaseEmbedder]: Configured model instance

    Raises:
        ValueError: If configuration is invalid or provider not supported

    Examples:
        >>> # Language model (auto-detected)
        >>> llm = get_model("openai/gpt-4o")
        >>> llm.generate("Hello!")

        >>> # Embedding model (auto-detected from name)
        >>> embedder = get_model("openai/text-embedding-3-small")
        >>> embedder.generate("Hello!")

        >>> # Explicit type override
        >>> embedder = get_model("openai", "custom-ft", model_type="embedding")

        >>> # With dimensions for embeddings
        >>> embedder = get_model("openai/text-embedding-3-small", dimensions=512)

        >>> # Explicit language type for type safety
        >>> llm = get_model("openai/gpt-4o", model_type="language")
    """
    # Extract values from config if provided
    if config:
        provider = provider or config.provider
        model_name = model_name or config.model_name
        api_key = api_key or config.api_key
        if isinstance(config, EmbeddingModelConfig):
            dimensions = dimensions or config.dimensions

    # Parse shorthand notation (e.g., "openai/gpt-4o")
    if provider and "/" in provider and model_name is None:
        prov, model = provider.split("/", 1)
        provider, model_name = prov, model

    # Resolve defaults (defaults are stored as full ids: provider/name)
    if model_type is not None:
        # Explicit type provided
        resolved_type = ModelType(model_type)
        if resolved_type == ModelType.LANGUAGE:
            if provider is None and model_name is None:
                provider, model_name = parse_model_id(DEFAULT_LANGUAGE_MODEL)
            else:
                provider = provider or DEFAULT_LANGUAGE_MODEL_PROVIDER
                if model_name is None:
                    full_id = DEFAULT_LANGUAGE_MODELS.get(provider)
                    model_name = model_name_from_id(full_id) if full_id else None
        elif resolved_type == ModelType.EMBEDDING:
            if provider is None and model_name is None:
                provider, model_name = parse_model_id(DEFAULT_EMBEDDING_MODEL)
            else:
                provider = provider or DEFAULT_EMBEDDING_MODEL_PROVIDER
                if model_name is None:
                    full_id = DEFAULT_EMBEDDING_MODELS.get(provider)
                    model_name = model_name_from_id(full_id) if full_id else None
    else:
        # Auto-detect type
        if provider is None and model_name is None:
            provider, model_name = parse_model_id(DEFAULT_LANGUAGE_MODEL)
        else:
            provider = provider or DEFAULT_LANGUAGE_MODEL_PROVIDER
            if model_name is None:
                full_id = DEFAULT_LANGUAGE_MODELS.get(provider)
                model_name = model_name_from_id(full_id) if full_id else None
        resolved_type = _classify_model(provider, model_name)

    # Validate provider exists
    if provider not in UNIFIED_MODEL_REGISTRY:
        available = ", ".join(sorted(UNIFIED_MODEL_REGISTRY.keys()))
        raise ValueError(f"Provider '{provider}' not supported. Available providers: {available}")

    # Look up factory in unified registry
    provider_factories = UNIFIED_MODEL_REGISTRY.get(provider, {})
    factory = provider_factories.get(resolved_type)

    if factory is None:
        supported_types = [t.value for t in provider_factories.keys()]
        raise ValueError(
            f"Provider '{provider}' does not support model type "
            f"'{resolved_type.value}'. Supported types: {supported_types}"
        )

    # Dispatch: _ProviderSpec uses generic creator, callables are invoked directly
    if isinstance(factory, _ProviderSpec):
        return _create_from_spec(factory, resolved_type, model_name, api_key, dimensions, **kwargs)
    elif resolved_type == ModelType.EMBEDDING:
        return factory(model_name, api_key, dimensions, **kwargs)
    else:
        return factory(model_name, api_key, **kwargs)


# =============================================================================
# Typed Factory Functions: get_language_model, get_embedding_model
# =============================================================================


def get_language_model(
    provider: Optional[str] = None,
    model_name: Optional[str] = None,
    api_key: Optional[str] = None,
    config: Optional[LanguageModelConfig] = None,
    **kwargs,
) -> BaseLLM:
    """Create a language model instance (LLM).

    Same as get_model() but always returns a language model, using language
    defaults when provider/model are omitted. Use this when you want a
    BaseLLM and do not want auto-detection.

    Args:
        provider: Provider name or "provider/model_name" as first positional
        model_name: Specific model name (omit when using unified format)
        api_key: API key for authentication
        config: Complete configuration object
        **kwargs: Additional parameters passed to the model

    Returns:
        BaseLLM: Configured language model instance

    Examples:
        >>> llm = get_language_model()
        >>> llm = get_language_model("openai/gpt-4o")
        >>> llm = get_language_model("anthropic", "claude-4")
    """
    return get_model(
        provider=provider,
        model_name=model_name,
        api_key=api_key,
        model_type=ModelType.LANGUAGE,
        config=config,
        **kwargs,
    )


def get_embedding_model(
    provider: Optional[str] = None,
    model_name: Optional[str] = None,
    api_key: Optional[str] = None,
    dimensions: Optional[int] = None,
    config: Optional[EmbeddingModelConfig] = None,
    **kwargs,
) -> BaseEmbedder:
    """Create an embedding model instance.

    Same as get_model(..., model_type="embedding") with embedding defaults.
    Use this when you want a BaseEmbedder without passing model_type.

    Args:
        provider: Provider name or "provider/model_name" as first positional
        model_name: Specific model name (omit when using unified format)
        api_key: API key for authentication
        dimensions: Optional embedding dimensions (model-dependent)
        config: Complete configuration object
        **kwargs: Additional parameters passed to the model

    Returns:
        BaseEmbedder: Configured embedding model instance

    Examples:
        >>> embedder = get_embedding_model()
        >>> embedder = get_embedding_model("openai/text-embedding-3-small")
        >>> embedder = get_embedding_model("openai", dimensions=512)
    """
    return get_model(
        provider=provider,
        model_name=model_name,
        api_key=api_key,
        model_type=ModelType.EMBEDDING,
        dimensions=dimensions,
        config=config,
        **kwargs,
    )


# Backward compatibility: get_embedder was the previous name for embedding models
get_embedder = get_embedding_model


# =============================================================================
# Helper Functions for Listing Available Models
# =============================================================================

# Providers that support listing available models, keyed by (module, class_name)
_LISTABLE_LLM_PROVIDERS: Dict[str, tuple[str, str]] = {
    "anthropic": (f"{_PROVIDERS_MODULE}.anthropic", "AnthropicLLM"),
    "cohere": (f"{_PROVIDERS_MODULE}.cohere", "CohereLLM"),
    "gemini": (f"{_PROVIDERS_MODULE}.gemini", "GeminiLLM"),
    "groq": (f"{_PROVIDERS_MODULE}.groq", "GroqLLM"),
    "meta_llama": (f"{_PROVIDERS_MODULE}.meta_llama", "MetaLlamaLLM"),
    "mistral": (f"{_PROVIDERS_MODULE}.mistral", "MistralLLM"),
    "ollama": (f"{_PROVIDERS_MODULE}.ollama", "OllamaLLM"),
    "openai": (f"{_PROVIDERS_MODULE}.openai", "OpenAILLM"),
    "openrouter": (f"{_PROVIDERS_MODULE}.openrouter", "OpenRouterLLM"),
    "perplexity": (f"{_PROVIDERS_MODULE}.perplexity", "PerplexityLLM"),
    "replicate": (f"{_PROVIDERS_MODULE}.replicate", "ReplicateLLM"),
    "together_ai": (f"{_PROVIDERS_MODULE}.together_ai", "TogetherAILLM"),
    "vertex_ai": (f"{_PROVIDERS_MODULE}.vertex_ai", "VertexAILLM"),
    "azure_ai": (f"{_PROVIDERS_MODULE}.azure_ai", "AzureAILLM"),
    "azure": (f"{_PROVIDERS_MODULE}.azure_openai", "AzureOpenAILLM"),
}

_LISTABLE_EMBEDDING_PROVIDERS: Dict[str, tuple[str, str]] = {
    "openai": (f"{_PROVIDERS_MODULE}.openai", "OpenAIEmbedder"),
    "gemini": (f"{_PROVIDERS_MODULE}.gemini", "GeminiEmbedder"),
    "vertex_ai": (f"{_PROVIDERS_MODULE}.vertex_ai", "VertexAIEmbedder"),
}


def _list_models(module_path: str, class_name: str) -> list[str]:
    """Import a provider class and call its get_available_models() class method."""
    mod = importlib.import_module(module_path)
    cls = getattr(mod, class_name)
    return cls.get_available_models()


def get_available_language_models(provider: str) -> list[str]:
    """Get the list of available language models for a specific provider.

    This function retrieves the available language models by calling the provider class's
    get_available_models() method. It supports all LiteLLM-based providers.

    Args:
        provider: Provider name (e.g., "anthropic", "openai", "gemini", "groq")

    Returns:
        List of available language model names for the provider

    Raises:
        ValueError: If the provider is not supported or doesn't support listing models
        ImportError: If required dependencies for the provider are missing

    Examples:
        >>> # Get Anthropic models
        >>> models = get_available_language_models("anthropic")
        >>> print(models)
        ['claude-3-5-sonnet-20241022', 'claude-3-5-haiku-20241022', ...]

        >>> # Get OpenAI models
        >>> models = get_available_language_models("openai")

        >>> # Get Gemini models
        >>> models = get_available_language_models("gemini")
    """
    if provider not in UNIFIED_MODEL_REGISTRY:
        available_providers = ", ".join(sorted(UNIFIED_MODEL_REGISTRY.keys()))
        raise ValueError(
            f"Provider '{provider}' not supported. Available providers: {available_providers}"
        )

    if ModelType.LANGUAGE not in UNIFIED_MODEL_REGISTRY.get(provider, {}):
        raise ValueError(f"Provider '{provider}' does not support language models.")

    if provider not in _LISTABLE_LLM_PROVIDERS:
        raise ValueError(
            f"Provider '{provider}' does not support listing "
            f"available models. Only the following providers support "
            f"this feature: "
            f"{', '.join(sorted(_LISTABLE_LLM_PROVIDERS.keys()))}"
        )

    module_path, class_name = _LISTABLE_LLM_PROVIDERS[provider]
    return _list_models(module_path, class_name)


def get_available_embedding_models(provider: str) -> list[str]:
    """Get the list of available embedding models for a specific provider.

    This function retrieves available embedding models by calling the provider's
    embedder class get_available_models() method. It supports OpenAI, Gemini,
    and Vertex AI providers.

    Args:
        provider: Provider name (e.g., "openai", "gemini", "vertex_ai")

    Returns:
        List of available embedding model names for the provider

    Raises:
        ValueError: If the provider is not supported or doesn't support embeddings
        ImportError: If required dependencies for the provider are missing

    Examples:
        >>> # Get OpenAI embedding models
        >>> models = get_available_embedding_models("openai")
        >>> print(models)
        ['text-embedding-3-small', 'text-embedding-3-large', 'text-embedding-ada-002']

        >>> # Get Gemini embedding models
        >>> models = get_available_embedding_models("gemini")

        >>> # Get Vertex AI embedding models
        >>> models = get_available_embedding_models("vertex_ai")
    """
    if provider not in UNIFIED_MODEL_REGISTRY:
        available_providers = ", ".join(sorted(UNIFIED_MODEL_REGISTRY.keys()))
        raise ValueError(
            f"Embedding provider '{provider}' not supported. "
            f"Available providers: {available_providers}"
        )

    if ModelType.EMBEDDING not in UNIFIED_MODEL_REGISTRY.get(provider, {}):
        raise ValueError(f"Provider '{provider}' does not support embedding models.")

    if provider not in _LISTABLE_EMBEDDING_PROVIDERS:
        raise ValueError(
            f"Provider '{provider}' does not support listing "
            f"available embedding models. Only the following providers "
            f"support this feature: "
            f"{', '.join(sorted(_LISTABLE_EMBEDDING_PROVIDERS.keys()))}"
        )

    module_path, class_name = _LISTABLE_EMBEDDING_PROVIDERS[provider]
    return _list_models(module_path, class_name)



import logging

import numpy as np

from .embedders import cos_sim
from .schemas import TopicNode

log = logging.getLogger(__name__)


def _is_under_topic(parent_path: str, child_path: str) -> bool:
    """Check if child_path is equal to or under parent_path.

    This is equivalent to the old is_subtopic() function.
    Returns True if child_path == parent_path or child_path is a descendant of parent_path.
    """
    if parent_path == child_path:
        return True
    if not parent_path:
        return True  # Everything is under root
    parent = TopicNode(path=parent_path)
    child = TopicNode(path=child_path)
    return parent.is_ancestor_of(child)


class PromptBuilder:
    """A class to build prompts for the model."""

    def __init__(
        self,
        prompt_size=7,
        slot_randomization=0.25,
        score_randomization=0.05,
        skip_randomization=0.25,
        prompt_diversity=True,
        subtopic_diversity=True,
    ):
        """Initialize the prompt builder.

        Parameters
        ----------
        prompt_size : int
            The number of test slots to include in the prompt.

        slot_randomization : float
            The proportion of slots to make fully random (within the current topic).

        score_randomization : float
            The standard deviation of an additive Gaussian randomization factor for the scores.

        skip_randomization : float
            The proportion of times we skip over top ranking tests when building the prompt.

        prompt_diversity : bool
            Whether to include a diversity term when selecting tests for the prompt.
            This diversity term is based on the embeddings of each test.

        subtopic_diversity : bool
            If true, we will try and pick tests from a diverse set of subtopics
            of the current topic (if we are using subtopic tests and not direct
            child tests).
        """

        assert skip_randomization < 0.99, (
            "skip_randomization must be less than 1, otherwise everything will always be skipped!"
        )

        self.prompt_size = prompt_size
        self.slot_randomization = slot_randomization
        # TODO: make this scale according to the stddev of the top 7 entries?
        self.score_randomization = score_randomization
        self.skip_randomization = skip_randomization
        self.prompt_diversity = prompt_diversity
        self.subtopic_diversity = subtopic_diversity

    def __call__(
        self,
        test_tree,
        topic,
        score_column,
        repetitions=1,
        filter="",
        suggest_topics=False,
        working_set_size=100,
        embeddings=None,
        embed_fn=None,
    ):
        """This builds a prompt for GPT3 that elicits useful input examples.

        Parameters
        ----------
        test_tree : adaptive_testing.TestTree
            The test tree to generate prompts from.

        topic : str
            The topic to build a prompt for.

        score_column : str
            The column to use for scoring the tests.

        repetitions : int
            The number of times to repeat the prompt generation process.
            This is how many prompts we will return.

        filter : str
            A filter to apply to the test set before selecting tests to build the prompt.

        suggest_topics : bool
            If true, we will create a prompt filled with topic names instead of a list of tests.

        working_set_size : int
            How many top tests to consider when doing the full iterative scoring
            process. Larger values may take longer. Note that this has no effect
            as long as we never go more than working_set_size tests deep during
            the prompt item selection process.

        embeddings : dict
            A dictionary of embeddings to use for the prompt.
            This is used to compute the prompt_diversity.
        """

        ids = np.array([node.id for node in test_tree])

        # return early for an empty test tree
        if len(ids) == 0:
            return [[] for _ in range(repetitions)]

        # we compute each test's distance from current topic, where distance is measured
        # by the length of the topic prefix shared between the test and the current topic
        topic_scaling = np.ones(len(test_tree))
        topic_parts = topic.split("/")
        for i in range(1, len(topic_parts)):
            prefix = "/".join(topic_parts[: i + 1])
            if suggest_topics:
                prefix += "/"
            topic_scaling *= 1 + 99 * np.array(
                [node.topic.startswith(prefix) for node in test_tree]
            )

        # promote direct children over subtopic descendants and filter for topics vs tests
        if suggest_topics:
            topic_scaling *= 1 + 99 * np.array(
                [node.topic.rsplit("/", 1)[0] == topic for node in test_tree]
            )
            topic_scaling *= np.array([node.label == "topic_marker" for node in test_tree])
        else:
            topic_scaling *= 1 + 99 * np.array([node.topic == topic for node in test_tree])
            topic_scaling *= np.array([node.label != "topic_marker" for node in test_tree])
        topic_scaling *= np.array(["__suggestions__" not in node.topic for node in test_tree])

        # return early if we have nothing to build a prompt with
        if np.sum(topic_scaling) == 0:
            return [[] for _ in range(repetitions)]

        topic_scaling /= np.max(topic_scaling)

        # filter down to a single test type (chosen to match the top scoring test)
        if suggest_topics:
            # scores currently do not influence topic suggestions
            # TODO: can we score topics and topic suggestions?
            scores = np.ones(len(ids))
        else:
            # compute a positive single value score for each test
            scores = np.array(
                [score_max(test_tree[k].model_score, test_tree[k].label) for k in ids]
            )

        # filter down to just top rows we will use during the iterative scoring process
        rank_vals = scores * topic_scaling
        top_inds = np.argsort(-rank_vals)[:working_set_size]
        ids = ids[top_inds]
        topic_scaling = topic_scaling[top_inds]
        scores = scores[top_inds] * 1.0

        # build a list of randomized prompts
        prompts = []
        for _ in range(repetitions):
            # store tmp versions of things we update during the iteration
            scores_curr = scores.copy()
            topic_scaling_curr = topic_scaling.copy()

            # score randomization
            scores_curr += self.score_randomization * np.random.rand(len(ids))

            # sim_avoidance is a vector that marks which items (and items related
            # through similarities) should be avoided (ranked lower for prompt selection)
            if self.prompt_diversity and embed_fn is not None:
                sim_avoidance = np.zeros(len(ids))
                if suggest_topics:
                    embeddings_arr = np.vstack(
                        embed_fn([test_tree[id].topic.split("/")[-1] for id in ids])
                    )
                else:
                    embeddings_arr = np.hstack(
                        [
                            np.vstack(embed_fn([test_tree[id].input for id in ids])),
                            np.vstack(embed_fn([test_tree[id].output for id in ids])),
                        ]
                    )
                similarities = cos_sim(embeddings_arr, embeddings_arr)
            elif self.prompt_diversity:
                sim_avoidance = np.zeros(len(ids))
                similarities = np.eye(len(ids))  # No embedder, skip similarity
            hard_avoidance = np.zeros(len(ids))
            diversity = np.ones(len(ids))

            # compute how many greedy and how many random positions we will have
            num_random = max(
                0,
                min(
                    np.random.binomial(self.prompt_size, self.slot_randomization),
                    len(ids) - self.prompt_size,
                ),
            )
            num_greedy = max(0, min(self.prompt_size - num_random, len(ids) - num_random))

            # iteratively select prompt items
            prompt_ids = []
            outside_topics_used = np.ones(len(ids))
            while len(prompt_ids) < num_greedy + num_random:
                # once we get to the random part of the process we scramble the scores
                if len(prompt_ids) == num_greedy:
                    scores_curr = 1 + np.random.rand(len(ids)) * 0.1

                # find the next bext index
                if self.prompt_diversity:
                    diversity = 1 - (similarities * sim_avoidance).max(1)
                rank_vals = (
                    scores_curr
                    * topic_scaling_curr
                    * diversity
                    * (1 - hard_avoidance)
                    * outside_topics_used
                )

                if (
                    np.nanmax(rank_vals) <= 0 and len(prompt_ids) > 0
                ):  # stop if we have run out of the current subtree
                    break

                new_ind = np.nanargmax(rank_vals)
                skip_rand = np.random.rand()

                # make it unlikely we will choose the same outside topic twice
                new_ind_topic = test_tree[ids[new_ind]].topic
                if not _is_under_topic(topic, new_ind_topic):
                    outside_topics_used *= 1 - 0.9 * np.array(
                        [test_tree[id].topic == new_ind_topic for id in ids]
                    )

                # add or skip this item
                if skip_rand >= self.skip_randomization:
                    prompt_ids.append(ids[new_ind])
                    avoidance_level = 1
                else:
                    avoidance_level = 1 - 0.1

                # avoid this IO pair as we select the next pairs
                hard_avoidance[new_ind] = avoidance_level
                if self.prompt_diversity:
                    sim_avoidance[new_ind] = avoidance_level

                # lower the weight of the subtopic we just picked from
                if self.subtopic_diversity:
                    new_topic = test_tree[ids[new_ind]].topic
                    if topic != new_topic and _is_under_topic(topic, new_topic):
                        subtopic = topic + "/" + new_topic[(len(topic) + 1) :].split("/")[0]
                        subtopic_scaling = np.array(
                            [
                                0.001 if _is_under_topic(subtopic, test_tree[k].topic) else 1
                                for k in ids
                            ]
                        )
                        topic_scaling_curr *= subtopic_scaling

            # create the prompt as a list of tuples
            prompt = []
            for k in reversed(prompt_ids):
                node = test_tree[k]
                if suggest_topics:
                    if node.topic == "":
                        continue  # we can't use the root to help suggest topic names
                    if "/" in node.topic:
                        parents, child = node.topic.rsplit("/", 1)
                    else:
                        parents, child = "", node.topic  # root-level topic
                    prompt.append((k, parents, child))
                else:
                    prompt.append((k, node.topic, node.input))
            prompts.append(prompt)

        return prompts


def score_max(s, label):
    if s == "" or s is None:
        return 1 if label == "fail" else 0
    elif isinstance(s, str):
        return np.max([convert_float(v) for v in s.split("|")])
    elif np.isnan(s):
        return 1 if label == "fail" else 0
    else:
        return np.max(s)


def convert_float(s):
    try:
        f = float(s)
    except ValueError:
        f = np.nan
    return f
